class NothingToDo(Exception):
    pass


class UsageError(Exception):
    pass


class UnsupportedProviderError(UsageError):
    def __init__(self, provider: str):
        super().__init__(
            "This provider is not supported: {p}".format(p=provider))
        self.provider = provider


class Error(Exception):
    pass


class ClusterNotFound(Error):
    pass


class ClusterAlreadyExists(Error):
    pass


class ClusterInvalidState(Error):
    def __init__(self, *, attempted_command: str, state: str):
        super().__init__(
            "Cluster is in state '{s}'. Cannot execute {c}.".format(
                c=attempted_command,
                s=state))
        self.attempted_command = attempted_command
        self.state = state


class SSHError(Error):
    def __init__(self, *, host: str, message: str):
        super().__init__(
            "[{h}] {m}".format(h=host, m=message))
        self.host = host
        self.message = message


class InterruptedEC2Operation(Error):
    def __init__(self, *, instances: list):
        super().__init__(
            "Operation aborted."
        )
        self.instances = instances



import os
import posixpath
import errno
import json
import resource
import sys
import shutil
import textwrap
import urllib.parse
import urllib.request
import warnings
import logging

# External modules
import click
import yaml
# We import botocore here so we can catch when the user tries to
# access AWS without having their credentials configured and provide
# a friendly error message. Apart from that, flintrock.py should
# not really know anything about EC2 or boto since that is delegated
# to ec2.py.
import botocore

# Flintrock modules
from . import ec2
from .exceptions import (
    UsageError,
    UnsupportedProviderError,
    NothingToDo,
    Error)
from flintrock import __version__
from .services import HDFS, Spark  # TODO: Remove this dependency.

FROZEN = getattr(sys, 'frozen', False)

if FROZEN:
    THIS_DIR = sys._MEIPASS
else:
    THIS_DIR = os.path.dirname(os.path.realpath(__file__))


logger = logging.getLogger('flintrock.flintrock')


def format_message(*, message: str, indent: int=4, wrap: int=70):
    """
    Format a lengthy message for printing to screen.
    """
    return textwrap.indent(
        textwrap.fill(
            textwrap.dedent(text=message),
            width=wrap),
        prefix=' ' * indent)


def option_name_to_variable_name(option: str):
    """
    Convert an option name like `--ec2-user` to the Python name it gets mapped to,
    like `ec2_user`.
    """
    return option.replace('--', '', 1).replace('-', '_')


def variable_name_to_option_name(variable: str):
    """
    Convert a variable name like `ec2_user` to the Click option name it gets mapped to,
    like `--ec2-user`.
    """
    return '--' + variable.replace('_', '-')


def option_requires(
        *,
        option: str,
        conditional_value=None,
        requires_all: list=[],
        requires_any: list=[],
        scope: dict):
    """
    Raise an exception if an option's requirements are not met.

    The option's requirements are checked only if the option has a "truthy" value
    (i.e. it's not a "falsy" value like '', None, or False), and if its value is
    equal to conditional_value, if conditional_value is not None.

    requires_all: Every option in this list must be defined.
    requires_any: At least one option in this list must be defined.

    This function looks for values by converting the option names to their
    corresponding variable names (e.g. --option-a becomes option_a) and looking them
    up in the provided scope.
    """
    option_value = scope[option_name_to_variable_name(option)]

    if option_value and \
            (conditional_value is None or option_value == conditional_value):
        if requires_all:
            for required_option in requires_all:
                required_name = option_name_to_variable_name(required_option)
                if required_name not in scope or not scope[required_name]:
                    raise UsageError(
                        "Error: Missing option \"{missing_option}\" is required by "
                        "\"{option}{space}{conditional_value}\"."
                        .format(
                            missing_option=required_option,
                            option=option,
                            space=' ' if conditional_value is not None else '',
                            conditional_value=conditional_value if conditional_value is not None else ''))
        if requires_any:
            for required_option in requires_any:
                required_name = option_name_to_variable_name(required_option)
                if required_name in scope and scope[required_name] is not None:
                    break
            else:
                raise UsageError(
                    "Error: \"{option}{space}{conditional_value}\" requires at least "
                    "one of the following options to be set: {at_least}"
                    .format(
                        option=option,
                        space=' ' if conditional_value is not None else '',
                        conditional_value=conditional_value if conditional_value is not None else '',
                        at_least=', '.join(['"' + ra + '"' for ra in requires_any])))


def mutually_exclusive(*, options: list, scope: dict):
    """
    Raise an exception if more than one of the provided options is specified.

    This function looks for values by converting the option names to their
    corresponding variable names (e.g. --option-a becomes option_a) and looking them
    up in the provided scope.
    """
    mutually_exclusive_names = [option_name_to_variable_name(o) for o in options]

    used_options = set()
    for name, value in scope.items():
        if name in mutually_exclusive_names and scope[name]:  # is not None:
            used_options.add(name)

    if len(used_options) > 1:
        bad_option1 = used_options.pop()
        bad_option2 = used_options.pop()
        raise UsageError(
            "Error: \"{option1}\" and \"{option2}\" are mutually exclusive.\n"
            "  {option1}: {value1}\n"
            "  {option2}: {value2}"
            .format(
                option1=variable_name_to_option_name(bad_option1),
                value1=scope[bad_option1],
                option2=variable_name_to_option_name(bad_option2),
                value2=scope[bad_option2]))


def get_config_file() -> str:
    """
    Get the path to Flintrock's default configuration file.
    """
    config_dir = click.get_app_dir(app_name='Flintrock')
    config_file = os.path.join(config_dir, 'config.yaml')
    return config_file


def configure_log(debug: bool):
    root_logger = logging.getLogger('flintrock')
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    if debug:
        root_logger.setLevel(logging.DEBUG)
        handler.setFormatter(logging.Formatter('%(asctime)s - flintrock.%(module)-9s - %(levelname)-5s - %(message)s'))
    else:
        root_logger.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter('%(message)s'))
    root_logger.addHandler(handler)


def build_hdfs_download_url(ctx, param, value):
    hdfs_download_url = value.format(v=ctx.params['hdfs_version'])
    return hdfs_download_url


def build_spark_download_url(ctx, param, value):
    spark_download_url = value.format(v=ctx.params['spark_version'])
    return spark_download_url


def validate_download_source(url):
    if 'spark' in url:
        software = 'Spark'
    elif 'hadoop' in url:
        software = 'Hadoop'
    else:
        software = 'software'

    parsed_url = urllib.parse.urlparse(url)

    if parsed_url.netloc == 'www.apache.org' and parsed_url.path == '/dyn/closer.lua':
        logger.warning(
            "Warning: "
            "Downloading {software} from an Apache mirror. Apache mirrors are "
            "often slow and unreliable, and typically only serve the most recent releases. "
            "We strongly recommend you specify a custom download source. "
            "For more background on this issue, please see: https://github.com/nchammas/flintrock/issues/238"
            .format(
                software=software,
            )
        )
        try:
            urllib.request.urlopen(url)
        except urllib.error.HTTPError as e:
            raise Error(
                "Error: Could not access {software} download. Maybe try a more recent release?\n"
                "  - Automatically redirected to: {url}\n"
                "  - HTTP error: {code}"
                .format(
                    software=software,
                    url=e.url,
                    code=e.code,
                )
            )


@click.group()
@click.option(
    '--config',
    help="Path to a Flintrock configuration file.",
    default=get_config_file())
@click.option('--provider', default='ec2', type=click.Choice(['ec2']))
@click.version_option(version=__version__)
# TODO: implement some solution like in https://github.com/pallets/click/issues/108
@click.option('--debug/--no-debug', default=False, help="Show debug information.")
@click.pass_context
def cli(cli_context, config, provider, debug):
    """
    Flintrock

    A command-line tool for launching Apache Spark clusters.
    """
    cli_context.obj['provider'] = provider

    if os.path.isfile(config):
        with open(config) as f:
            config_raw = yaml.safe_load(f)
            debug = config_raw.get('debug') or debug
            config_map = config_to_click(normalize_keys(config_raw))

        cli_context.default_map = config_map
    else:
        if config != get_config_file():
            raise FileNotFoundError(errno.ENOENT, 'No such file', config)
    configure_log(debug=debug)


@cli.command()
@click.argument('cluster-name')
@click.option('--num-slaves', type=click.IntRange(min=1), required=True)
@click.option('--install-hdfs/--no-install-hdfs', default=False)
@click.option('--hdfs-version', default='2.8.5')
@click.option('--hdfs-download-source',
              help="URL to download Hadoop from.",
              default='https://www.apache.org/dyn/closer.lua?action=download&filename=hadoop/common/hadoop-{v}/hadoop-{v}.tar.gz',
              show_default=True,
              callback=build_hdfs_download_url)
@click.option('--install-spark/--no-install-spark', default=True)
@click.option('--spark-executor-instances', default=1,
              help="How many executor instances per worker.")
@click.option('--spark-version',
              # Don't set a default here because it will conflict with
              # the config file if the git commit is set.
              # See: https://github.com/nchammas/flintrock/issues/190
              # default=,
              help="Spark release version to install.")
@click.option('--spark-download-source',
              help="URL to download a release of Spark from.",
              default='https://www.apache.org/dyn/closer.lua?action=download&filename=spark/spark-{v}/spark-{v}-bin-hadoop2.7.tgz',
              show_default=True,
              callback=build_spark_download_url)
@click.option('--spark-git-commit',
              help="Git commit to build Spark from. "
                   "Set to 'latest' to build Spark from the latest commit on the "
                   "repository's default branch.")
@click.option('--spark-git-repository',
              help="Git repository to clone Spark from.",
              default='https://github.com/apache/spark',
              show_default=True)
@click.option('--assume-yes/--no-assume-yes', default=False)
@click.option('--ec2-key-name')
@click.option('--ec2-identity-file',
              type=click.Path(exists=True, dir_okay=False),
              help="Path to SSH .pem file for accessing nodes.")
@click.option('--ec2-instance-type', default='m3.medium', show_default=True)
@click.option('--ec2-region', default='us-east-1', show_default=True)
# We set some of these defaults to empty strings because of boto3's parameter validation.
# See: https://github.com/boto/boto3/issues/400
@click.option('--ec2-availability-zone', default='')
@click.option('--ec2-ami')
@click.option('--ec2-user')
@click.option('--ec2-security-group', 'ec2_security_groups',
              multiple=True,
              help="Additional security groups names to assign to the instances. "
                   "You can specify this option multiple times.")
@click.option('--ec2-spot-price', type=float)
@click.option('--ec2-min-root-ebs-size-gb', type=int, default=30)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
@click.option('--ec2-subnet-id', default='')
@click.option('--ec2-instance-profile-name', default='')
@click.option('--ec2-placement-group', default='')
@click.option('--ec2-tenancy', default='default')
@click.option('--ec2-ebs-optimized/--no-ec2-ebs-optimized', default=False)
@click.option('--ec2-instance-initiated-shutdown-behavior', default='stop',
              type=click.Choice(['stop', 'terminate']))
@click.option('--ec2-user-data',
              type=click.File(mode='r', encoding='utf-8'),
              help="Path to EC2 user data script that will run on instance launch.")
@click.option('--ec2-tag', 'ec2_tags',
              callback=ec2.cli_validate_tags,
              multiple=True,
              help="Additional tags (e.g. 'Key,Value') to assign to the instances. "
                   "You can specify this option multiple times.")
@click.pass_context
def launch(
        cli_context,
        cluster_name,
        num_slaves,
        install_hdfs,
        hdfs_version,
        hdfs_download_source,
        install_spark,
        spark_executor_instances,
        spark_version,
        spark_git_commit,
        spark_git_repository,
        spark_download_source,
        assume_yes,
        ec2_key_name,
        ec2_identity_file,
        ec2_instance_type,
        ec2_region,
        ec2_availability_zone,
        ec2_ami,
        ec2_user,
        ec2_security_groups,
        ec2_spot_price,
        ec2_min_root_ebs_size_gb,
        ec2_vpc_id,
        ec2_subnet_id,
        ec2_instance_profile_name,
        ec2_placement_group,
        ec2_tenancy,
        ec2_ebs_optimized,
        ec2_instance_initiated_shutdown_behavior,
        ec2_user_data,
        ec2_tags):
    """
    Launch a new cluster.
    """
    provider = cli_context.obj['provider']
    services = []

    option_requires(
        option='--install-hdfs',
        requires_all=['--hdfs-version'],
        scope=locals())
    option_requires(
        option='--install-spark',
        requires_any=[
            '--spark-version',
            '--spark-git-commit'],
        scope=locals())
    mutually_exclusive(
        options=[
            '--spark-version',
            '--spark-git-commit'],
        scope=locals())
    option_requires(
        option='--install-spark',
        requires_all=[
            '--hdfs-version'],
        scope=locals())
    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=[
            '--ec2-key-name',
            '--ec2-identity-file',
            '--ec2-instance-type',
            '--ec2-region',
            '--ec2-ami',
            '--ec2-user'],
        scope=locals())
    # The subnet is required for non-default VPCs because EC2 does not
    # support user-defined default subnets.
    # See: https://forums.aws.amazon.com/thread.jspa?messageID=707417
    #      https://github.com/mitchellh/packer/issues/1935#issuecomment-111235752
    option_requires(
        option='--ec2-vpc-id',
        requires_all=['--ec2-subnet-id'],
        scope=locals())

    check_external_dependency('ssh-keygen')

    if install_hdfs:
        validate_download_source(hdfs_download_source)
        hdfs = HDFS(
            version=hdfs_version,
            download_source=hdfs_download_source,
        )
        services += [hdfs]
    if install_spark:
        if spark_version:
            validate_download_source(spark_download_source)
            spark = Spark(
                spark_executor_instances=spark_executor_instances,
                version=spark_version,
                hadoop_version=hdfs_version,
                download_source=spark_download_source,
            )
        elif spark_git_commit:
            logger.warning(
                "Warning: Building Spark takes a long time. "
                "e.g. 15-20 minutes on an m3.xlarge instance on EC2.")
            if spark_git_commit == 'latest':
                spark_git_commit = get_latest_commit(spark_git_repository)
                logger.info("Building Spark at latest commit: {c}".format(c=spark_git_commit))
            spark = Spark(
                spark_executor_instances=spark_executor_instances,
                git_commit=spark_git_commit,
                git_repository=spark_git_repository,
                hadoop_version=hdfs_version,
            )
        services += [spark]

    if provider == 'ec2':
        cluster = ec2.launch(
            cluster_name=cluster_name,
            num_slaves=num_slaves,
            services=services,
            assume_yes=assume_yes,
            key_name=ec2_key_name,
            identity_file=ec2_identity_file,
            instance_type=ec2_instance_type,
            region=ec2_region,
            availability_zone=ec2_availability_zone,
            ami=ec2_ami,
            user=ec2_user,
            security_groups=ec2_security_groups,
            spot_price=ec2_spot_price,
            min_root_ebs_size_gb=ec2_min_root_ebs_size_gb,
            vpc_id=ec2_vpc_id,
            subnet_id=ec2_subnet_id,
            instance_profile_name=ec2_instance_profile_name,
            placement_group=ec2_placement_group,
            tenancy=ec2_tenancy,
            ebs_optimized=ec2_ebs_optimized,
            instance_initiated_shutdown_behavior=ec2_instance_initiated_shutdown_behavior,
            user_data=ec2_user_data,
            tags=ec2_tags)
    else:
        raise UnsupportedProviderError(provider)

    print("Cluster master: {}".format(cluster.master_host))
    print("Login with: flintrock login {}".format(cluster.name))


def get_latest_commit(github_repository: str):
    """
    Get the latest commit on the default branch of a repository hosted on GitHub.
    """
    parsed_url = urllib.parse.urlparse(github_repository)
    repo_domain, repo_path = parsed_url.netloc, parsed_url.path.strip('/')

    if repo_domain != 'github.com':
        raise UsageError(
            "Error: Getting the latest commit is only supported "
            "for repositories hosted on GitHub. "
            "Provided repository domain was: {d}".format(d=repo_domain))

    url = "https://api.github.com/repos/{rp}/commits".format(rp=repo_path)
    try:
        with urllib.request.urlopen(url) as response:
            result = json.loads(response.read().decode('utf-8'))
            return result[0]['sha']
    except Exception as e:
        raise Exception(
            "Could not get latest commit for repository: {r}"
            .format(r=repo_path)) from e


@cli.command()
@click.argument('cluster-name')
@click.option('--assume-yes/--no-assume-yes', default=False)
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
@click.pass_context
def destroy(cli_context, cluster_name, assume_yes, ec2_region, ec2_vpc_id):
    """
    Destroy a cluster.
    """
    provider = cli_context.obj['provider']

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=['--ec2-region'],
        scope=locals())

    if provider == 'ec2':
        cluster = ec2.get_cluster(
            cluster_name=cluster_name,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
    else:
        raise UnsupportedProviderError(provider)

    if not assume_yes:
        cluster.print()
        click.confirm(
            text="Are you sure you want to destroy this cluster?",
            abort=True)

    logger.info("Destroying {c}...".format(c=cluster.name))
    cluster.destroy()


@cli.command()
@click.argument('cluster-name', required=False)
@click.option('--master-hostname-only', is_flag=True, default=False)
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
@click.pass_context
def describe(
        cli_context,
        cluster_name,
        master_hostname_only,
        ec2_region,
        ec2_vpc_id):
    """
    Describe an existing cluster.

    Leave out the cluster name to find all Flintrock-managed clusters.

    The output of this command is both human- and machine-friendly. Full cluster
    descriptions are output in YAML.
    """
    provider = cli_context.obj['provider']
    search_area = ""

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=['--ec2-region'],
        scope=locals())

    if cluster_name:
        cluster_names = [cluster_name]
    else:
        cluster_names = []

    if provider == 'ec2':
        search_area = "in region {r}".format(r=ec2_region)
        clusters = ec2.get_clusters(
            cluster_names=cluster_names,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
    else:
        raise UnsupportedProviderError(provider)

    if cluster_name:
        cluster = clusters[0]
        if master_hostname_only:
            logger.info(cluster.master_host)
        else:
            cluster.print()
    else:
        if master_hostname_only:
            for cluster in sorted(clusters, key=lambda x: x.name):
                logger.info("{}: {}".format(cluster.name, cluster.master_host))
        else:
            logger.info("Found {n} cluster{s}{space}{search_area}.".format(
                n=len(clusters),
                s='' if len(clusters) == 1 else 's',
                space=' ' if search_area else '',
                search_area=search_area))
            if clusters:
                logger.info('---')
                for cluster in sorted(clusters, key=lambda x: x.name):
                    cluster.print()


# TODO: Provide different command or option for going straight to Spark Shell. (?)
@cli.command()
@click.argument('cluster-name')
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
# TODO: Move identity-file to global, non-provider-specific option. (?)
@click.option('--ec2-identity-file',
              type=click.Path(exists=True, dir_okay=False),
              help="Path to SSH .pem file for accessing nodes.")
@click.option('--ec2-user')
@click.pass_context
def login(cli_context, cluster_name, ec2_region, ec2_vpc_id, ec2_identity_file, ec2_user):
    """
    Login to the master of an existing cluster.
    """
    provider = cli_context.obj['provider']

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=[
            '--ec2-region',
            '--ec2-identity-file',
            '--ec2-user'],
        scope=locals())

    check_external_dependency('ssh')

    if provider == 'ec2':
        cluster = ec2.get_cluster(
            cluster_name=cluster_name,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
        user = ec2_user
        identity_file = ec2_identity_file
    else:
        raise UnsupportedProviderError(provider)

    # TODO: Check that master up first and error out cleanly if not
    #       via ClusterInvalidState.
    cluster.login(user=user, identity_file=identity_file)


@cli.command()
@click.argument('cluster-name')
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
# TODO: Move identity-file to global, non-provider-specific option. (?)
@click.option('--ec2-identity-file',
              type=click.Path(exists=True, dir_okay=False),
              help="Path to SSH .pem file for accessing nodes.")
@click.option('--ec2-user')
@click.pass_context
def start(cli_context, cluster_name, ec2_region, ec2_vpc_id, ec2_identity_file, ec2_user):
    """
    Start an existing, stopped cluster.
    """
    provider = cli_context.obj['provider']

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=[
            '--ec2-region',
            '--ec2-identity-file',
            '--ec2-user'],
        scope=locals())

    if provider == 'ec2':
        cluster = ec2.get_cluster(
            cluster_name=cluster_name,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
        user = ec2_user
        identity_file = ec2_identity_file
    else:
        raise UnsupportedProviderError(provider)

    cluster.start_check()
    logger.info("Starting {c}...".format(c=cluster_name))
    cluster.start(user=user, identity_file=identity_file)


@cli.command()
@click.argument('cluster-name')
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
@click.option('--assume-yes/--no-assume-yes', default=False)
@click.pass_context
def stop(cli_context, cluster_name, ec2_region, ec2_vpc_id, assume_yes):
    """
    Stop an existing, running cluster.
    """
    provider = cli_context.obj['provider']

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=['--ec2-region'],
        scope=locals())

    if provider == 'ec2':
        cluster = ec2.get_cluster(
            cluster_name=cluster_name,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
    else:
        raise UnsupportedProviderError(provider)

    cluster.stop_check()

    if not assume_yes:
        cluster.print()
        click.confirm(
            text="Are you sure you want to stop this cluster?",
            abort=True)

    logger.info("Stopping {c}...".format(c=cluster_name))
    cluster.stop()
    logger.info("{c} is now stopped.".format(c=cluster_name))


@cli.command(name='add-slaves')
@click.argument('cluster-name')
@click.option('--num-slaves', type=click.IntRange(min=1), required=True)
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
@click.option('--ec2-identity-file',
              type=click.Path(exists=True, dir_okay=False),
              help="Path to SSH .pem file for accessing nodes.")
@click.option('--ec2-user')
@click.option('--ec2-spot-price', type=float)
@click.option('--ec2-min-root-ebs-size-gb', type=int, default=30)
@click.option('--assume-yes/--no-assume-yes', default=False)
@click.option('--ec2-tag', 'ec2_tags',
              callback=ec2.cli_validate_tags,
              multiple=True,
              help="Additional tags (e.g. 'Key,Value') to assign to the instances. "
                   "You can specify this option multiple times.")
@click.pass_context
def add_slaves(
        cli_context,
        cluster_name,
        num_slaves,
        ec2_region,
        ec2_vpc_id,
        ec2_identity_file,
        ec2_user,
        ec2_spot_price,
        ec2_min_root_ebs_size_gb,
        ec2_tags,
        assume_yes):
    """
    Add slaves to an existing cluster.

    Flintrock will configure new slaves based on information queried
    automatically from the master.
    """
    provider = cli_context.obj['provider']

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=[
            '--ec2-region',
            '--ec2-identity-file',
            '--ec2-user'],
        scope=locals())

    if provider == 'ec2':
        cluster = ec2.get_cluster(
            cluster_name=cluster_name,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
        user = ec2_user
        identity_file = ec2_identity_file
        provider_options = {
            'min_root_ebs_size_gb': ec2_min_root_ebs_size_gb,
            'spot_price': ec2_spot_price,
            'tags': ec2_tags
        }
    else:
        raise UnsupportedProviderError(provider)

    if cluster.num_masters == 0:
        raise Error(
            "Cannot add slaves to cluster '{c}' since it does not "
            "appear to have a master."
            .format(
                c=cluster_name))

    cluster.load_manifest(
        user=user,
        identity_file=identity_file)
    cluster.add_slaves_check()

    if provider == 'ec2':
        cluster.add_slaves(
            user=user,
            identity_file=identity_file,
            num_slaves=num_slaves,
            assume_yes=assume_yes,
            **provider_options)


@cli.command(name='remove-slaves')
@click.argument('cluster-name')
@click.option('--num-slaves', type=click.IntRange(min=1), required=True)
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
@click.option('--ec2-user')
@click.option('--ec2-identity-file',
              type=click.Path(exists=True, dir_okay=False),
              help="Path to SSH .pem file for accessing nodes.")
@click.option('--assume-yes/--no-assume-yes', default=False)
@click.pass_context
def remove_slaves(
        cli_context,
        cluster_name,
        num_slaves,
        ec2_region,
        ec2_vpc_id,
        ec2_user,
        ec2_identity_file,
        assume_yes):
    """
    Remove slaves from an existing cluster.
    """
    provider = cli_context.obj['provider']

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=[
            '--ec2-region',
            '--ec2-user',
            '--ec2-identity-file'],
        scope=locals())

    if provider == 'ec2':
        cluster = ec2.get_cluster(
            cluster_name=cluster_name,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
        user = ec2_user
        identity_file = ec2_identity_file
    else:
        raise UnsupportedProviderError(provider)

    if num_slaves > cluster.num_slaves:
        logger.warning(
            "Warning: Cluster has {c} slave{cs}. "
            "You asked to remove {n} slave{ns}."
            .format(
                c=cluster.num_slaves,
                cs='' if cluster.num_slaves == 1 else 's',
                n=num_slaves,
                ns='' if num_slaves == 1 else 's'))
        num_slaves = cluster.num_slaves

    if not assume_yes:
        cluster.print()
        click.confirm(
            text=("Are you sure you want to remove {n} slave{s} from this cluster?"
                  .format(
                      n=num_slaves,
                      s='' if num_slaves == 1 else 's')),
            abort=True)

    logger.info("Removing {n} slave{s}..."
                .format(
                    n=num_slaves,
                    s='' if num_slaves == 1 else 's'))
    cluster.remove_slaves(
        user=user,
        identity_file=identity_file,
        num_slaves=num_slaves)


@cli.command(name='run-command')
@click.argument('cluster-name')
@click.argument('command', nargs=-1)
@click.option('--master-only', help="Run on the master only.", is_flag=True)
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
@click.option('--ec2-identity-file',
              type=click.Path(exists=True, dir_okay=False),
              help="Path to SSH .pem file for accessing nodes.")
@click.option('--ec2-user')
@click.pass_context
def run_command(
        cli_context,
        cluster_name,
        command,
        master_only,
        ec2_region,
        ec2_vpc_id,
        ec2_identity_file,
        ec2_user):
    """
    Run a shell command on a cluster.

    Examples:

        flintrock run-command my-cluster 'touch /tmp/flintrock'
        flintrock run-command my-cluster -- yum install -y package

    Flintrock will return a non-zero code if any of the cluster nodes raises an error
    while running the command.
    """
    provider = cli_context.obj['provider']

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=[
            '--ec2-region',
            '--ec2-identity-file',
            '--ec2-user'],
        scope=locals())

    if provider == 'ec2':
        cluster = ec2.get_cluster(
            cluster_name=cluster_name,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
        user = ec2_user
        identity_file = ec2_identity_file
    else:
        raise UnsupportedProviderError(provider)

    cluster.run_command_check()

    logger.info("Running command on {target}...".format(
        target="master only" if master_only else "cluster"))

    cluster.run_command(
        command=command,
        master_only=master_only,
        user=user,
        identity_file=identity_file)


@cli.command(name='copy-file')
@click.argument('cluster-name')
@click.argument('local_path', type=click.Path(exists=True, dir_okay=False))
@click.argument('remote_path', type=click.Path())
@click.option('--master-only', help="Copy to the master only.", is_flag=True)
@click.option('--ec2-region', default='us-east-1', show_default=True)
@click.option('--ec2-vpc-id', default='', help="Leave empty for default VPC.")
@click.option('--ec2-identity-file',
              type=click.Path(exists=True, dir_okay=False),
              help="Path to SSH .pem file for accessing nodes.")
@click.option('--ec2-user')
@click.option('--assume-yes/--no-assume-yes', default=False, help="Prompt before large uploads.")
@click.pass_context
def copy_file(
        cli_context,
        cluster_name,
        local_path,
        remote_path,
        master_only,
        ec2_region,
        ec2_vpc_id,
        ec2_identity_file,
        ec2_user,
        assume_yes):
    """
    Copy a local file up to a cluster.

    This will copy the file to the same path on each node of the cluster.

    Examples:

        flintrock copy-file my-cluster /tmp/file.102.txt /tmp/file.txt
        flintrock copy-file my-cluster /tmp/spark-defaults.conf /tmp/

    Flintrock will return a non-zero code if any of the cluster nodes raises an error.
    """
    provider = cli_context.obj['provider']

    option_requires(
        option='--provider',
        conditional_value='ec2',
        requires_all=[
            '--ec2-region',
            '--ec2-identity-file',
            '--ec2-user'],
        scope=locals())

    # We assume POSIX for the remote path since Flintrock
    # only supports clusters running CentOS / Amazon Linux.
    if not posixpath.basename(remote_path):
        remote_path = posixpath.join(remote_path, os.path.basename(local_path))

    if provider == 'ec2':
        cluster = ec2.get_cluster(
            cluster_name=cluster_name,
            region=ec2_region,
            vpc_id=ec2_vpc_id)
        user = ec2_user
        identity_file = ec2_identity_file
    else:
        raise UnsupportedProviderError(provider)

    cluster.copy_file_check()

    if not assume_yes and not master_only:
        file_size_bytes = os.path.getsize(local_path)
        num_nodes = len(cluster.slave_ips) + 1  # TODO: cluster.num_nodes
        total_size_bytes = file_size_bytes * num_nodes

        if total_size_bytes > 10 ** 6:
            logger.warning("WARNING:")
            logger.warning(
                format_message(
                    message="""\
                        You are trying to upload {total_size} bytes ({size} bytes x {count}
                        nodes in {cluster}). Depending on your upload bandwidth, this may take
                        a long time.
                        You may be better off uploading this file to a storage service like
                        Amazon S3 and downloading it from there to the cluster using
                        `flintrock run-command ...`.
                        """.format(
                            size=file_size_bytes,
                            count=num_nodes,
                            cluster=cluster_name,
                            total_size=total_size_bytes),
                    wrap=60))
            click.confirm(
                text="Are you sure you want to continue?",
                default=True,
                abort=True)

    logger.info("Copying file to {target}...".format(
        target="master only" if master_only else "cluster"))

    cluster.copy_file(
        local_path=local_path,
        remote_path=remote_path,
        master_only=master_only,
        user=user,
        identity_file=identity_file)


def normalize_keys(obj):
    """
    Used to map keys from config files to Python parameter names.
    """
    if type(obj) != dict:
        return obj
    else:
        return {k.replace('-', '_'): normalize_keys(v) for k, v in obj.items()}


def config_to_click(config: dict) -> dict:
    """
    Convert a dictionary of configurations loaded from a Flintrock config file
    to a dictionary that Click can use to set default options.
    """
    service_configs = {}

    if 'services' in config:
        for service in config['services']:
            if config['services'][service]:
                service_configs.update(
                    {service + '_' + k: v for (k, v) in config['services'][service].items()})

    ec2_configs = {
        'ec2_' + k: v for (k, v) in config['providers']['ec2'].items()}

    click_map = {
        'launch': dict(
            list(config['launch'].items()) +
            list(ec2_configs.items()) +
            list(service_configs.items())),
        'describe': ec2_configs,
        'destroy': ec2_configs,
        'login': ec2_configs,
        'start': ec2_configs,
        'stop': ec2_configs,
        'add-slaves': ec2_configs,
        'remove-slaves': ec2_configs,
        'run-command': ec2_configs,
        'copy-file': ec2_configs,
    }

    return click_map


@cli.command()
@click.option('--locate', is_flag=True, default=False,
              help="Don't open an editor. "
              "Just open the folder containing the configuration file.")
@click.pass_context
def configure(cli_context, locate):
    """
    Configure Flintrock's defaults.

    This will open Flintrock's configuration file in your default YAML editor so
    you can set your defaults.
    """
    config_file = get_config_file()

    if not os.path.isfile(config_file):
        logger.info("Initializing config file from template...")
        os.makedirs(os.path.dirname(config_file), exist_ok=True)
        shutil.copyfile(
            src=os.path.join(THIS_DIR, 'config.yaml.template'),
            dst=config_file)
        os.chmod(config_file, mode=0o644)

    ret = click.launch(config_file, locate=locate)

    if ret != 0:
        raise Error(
            "Flintrock could not launch an application to {action} "
            "the config file at '{location}'. You may want to manually "
            "find and edit this file."
            .format(
                action="locate" if locate else "edit",
                location=config_file
            )
        )


def flintrock_is_in_development_mode() -> bool:
    """
    Check if Flintrock was installed in development mode.

    Use this function to toggle behavior that only Flintrock developers should
    see.
    """
    # This esoteric technique was pulled from pip.
    # See: https://github.com/pypa/pip/pull/3258/files#diff-ab583908279e865537dec218246edcfcR310
    for path_item in sys.path:
        egg_link = os.path.join(path_item, 'Flintrock.egg-link')
        if os.path.isfile(egg_link):
            return True
    else:
        return False


def set_open_files_limit(desired_limit):
    """
    On POSIX systems, set the open files limit to the desired number, unless
    it is already equal to or higher than that.

    Setting a high limit enables Flintrock to launch or interact with really
    large clusters.

    Background discussion: https://github.com/nchammas/flintrock/issues/81
    """
    soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)

    if soft_limit < desired_limit:
        if desired_limit > hard_limit:
            warnings.warn(
                "Flintrock cannot set the open files limit to {desired} "
                "because the OS hard limit is {hard}. Going with {hard}. "
                "You may have problems launching or interacting with "
                "really large clusters."
                .format(
                    desired=desired_limit,
                    hard=hard_limit),
                category=RuntimeWarning,
                stacklevel=2)
        resource.setrlimit(
            resource.RLIMIT_NOFILE,
            (min(desired_limit, hard_limit), hard_limit))


def check_external_dependency(executable_name: str):
    if shutil.which(executable_name) is None:
        raise Error(
            "Error: Flintrock could not find '{executable}' on your PATH. "
            "Flintrock needs this executable to carry out the operation you "
            "requested. Please install it and try again."
            .format(
                executable=executable_name
            )
        )


def main() -> int:
    # Starting in Python 3.7, deprecation warnings are shown by default. We
    # don't want to show these to end-users.
    # See: https://docs.python.org/3/library/warnings.html#default-warning-filter
    if not flintrock_is_in_development_mode():
        warnings.simplefilter(action='ignore', category=DeprecationWarning)

    set_open_files_limit(4096)

    try:
        try:
            # We pass in obj so we can add attributes to it, like provider, which
            # get shared by all commands.
            # See: http://click.pocoo.org/6/api/#click.Context
            cli(obj={})
        except botocore.exceptions.NoCredentialsError:
            raise Error(
                "Flintrock could not find your AWS credentials. "
                "You can fix this is by providing your credentials "
                "via environment variables or by creating a shared "
                "credentials file.\n"
                "For more information see:\n"
                "  * https://boto3.readthedocs.io/en/latest/guide/configuration.html#environment-variables\n"
                "  * https://boto3.readthedocs.io/en/latest/guide/configuration.html#shared-credentials-file"
            )
    except NothingToDo as e:
        print(e)
        return 0
    except UsageError as e:
        print(e, file=sys.stderr)
        return 2
    except Error as e:
        print(e, file=sys.stderr)
        return 1



import os
import sys

FROZEN = getattr(sys, 'frozen', False)


def get_subprocess_env() -> dict:
    """
    Get the environment we want to use when making subprocess calls.
    This takes care of details that affect subprocess calls made from
    PyInstaller-packaged versions of Flintrock.

    For more information see: https://github.com/pyinstaller/pyinstaller/blob/develop/doc/runtime-information.rst#ld_library_path--libpath-considerations
    """
    env = dict(os.environ)
    if FROZEN:
        env['LD_LIBRARY_PATH'] = env.get('LD_LIBRARY_PATH_ORIG', '')
    return env



import errno
import os
import socket
import subprocess
import tempfile
import time
import logging
from collections import namedtuple

# External modules
import paramiko

# Flintrock modules
from .util import get_subprocess_env
from .exceptions import SSHError

SSHKeyPair = namedtuple('KeyPair', ['public', 'private'])


logger = logging.getLogger('flintrock.ssh')


def generate_ssh_key_pair() -> SSHKeyPair:
    """
    Generate an SSH key pair that the cluster can use for intra-cluster
    communication.
    """
    with tempfile.TemporaryDirectory() as tempdir:
        subprocess.check_call(
            [
                'ssh-keygen',
                '-q',
                '-t', 'rsa',
                '-N', '',
                '-f', os.path.join(tempdir, 'flintrock_rsa'),
                '-C', 'flintrock',
            ],
            env=get_subprocess_env(),
        )

        with open(file=os.path.join(tempdir, 'flintrock_rsa')) as private_key_file:
            private_key = private_key_file.read()

        with open(file=os.path.join(tempdir, 'flintrock_rsa.pub')) as public_key_file:
            public_key = public_key_file.read()

    return namedtuple('KeyPair', ['public', 'private'])(public_key, private_key)


def get_ssh_client(
        *,
        user: str,
        host: str,
        identity_file: str,
        wait: bool=False,
        print_status: bool=None) -> paramiko.client.SSHClient:
    """
    Get an SSH client for the provided host, waiting as necessary for SSH to become
    available.
    """
    if print_status is None:
        print_status = wait

    client = paramiko.client.SSHClient()

    client.load_system_host_keys()
    client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy())

    if wait:
        tries = 100
    else:
        # It's greater than 1 as a band-aid for this issue:
        # https://github.com/nchammas/flintrock/issues/198
        tries = 3

    while tries > 0:
        try:
            tries -= 1
            client.connect(
                username=user,
                hostname=host,
                key_filename=identity_file,
                look_for_keys=False,
                timeout=3)
            if print_status:
                logger.info("[{h}] SSH online.".format(h=host))
            break
        except socket.timeout as e:
            logger.debug("[{h}] SSH timeout.".format(h=host))
            time.sleep(5)
        except paramiko.ssh_exception.NoValidConnectionsError as e:
            if any(error.errno != errno.ECONNREFUSED for error in e.errors.values()):
                raise
            logger.debug("[{h}] SSH exception: {e}".format(h=host, e=e))
            time.sleep(5)
        # We get this exception during startup with CentOS but not Amazon Linux,
        # for some reason.
        except paramiko.ssh_exception.AuthenticationException as e:
            logger.debug("[{h}] SSH AuthenticationException.".format(h=host))
            time.sleep(5)
        except paramiko.ssh_exception.SSHException as e:
            raise SSHError(
                host=host,
                message="SSH protocol error. Possible causes include using "
                "the wrong key file or username.",
            ) from e
    else:
        raise SSHError(
            host=host,
            message="Could not connect via SSH.")

    return client


def ssh_check_output(
        client: paramiko.client.SSHClient,
        command: str,
        timeout_seconds: int=None,
):
    """
    Run a command via the provided SSH client and return the output captured
    on stdout.

    Raise an exception if the command returns a non-zero code.
    """
    stdin, stdout, stderr = client.exec_command(
        command,
        get_pty=True,
        timeout=timeout_seconds)

    # NOTE: Paramiko doesn't clearly document this, but we must read() before
    #       calling recv_exit_status().
    #       See: https://github.com/paramiko/paramiko/issues/448#issuecomment-159481997
    stdout_output = stdout.read().decode('utf8').rstrip('\n')
    stderr_output = stderr.read().decode('utf8').rstrip('\n')
    exit_status = stdout.channel.recv_exit_status()

    if exit_status:
        # TODO: Return a custom exception that includes the return code.
        #       See: https://docs.python.org/3/library/subprocess.html#subprocess.check_output
        # NOTE: We are losing the output order here since output from stdout and stderr
        #       may be interleaved.
        raise SSHError(
            host=client.get_transport().getpeername()[0],
            message=stdout_output + stderr_output)

    return stdout_output


def ssh(*, user: str, host: str, identity_file: str):
    """
    SSH into a host for interactive use.
    """
    subprocess.call(
        [
            'ssh',
            '-o', 'StrictHostKeyChecking=no',
            '-i', identity_file,
            '{u}@{h}'.format(u=user, h=host),
        ],
        env=get_subprocess_env(),
    )



"""
Setup ephemeral storage on a newly launched Linux host.

This script was developed against EC2, where ephemeral volumes are by
default haphazardly and inconsistently mounted. Therefore, we unmount
all volumes that we detect and explicitly format and remount them as
we please.

The resulting structure we create is as follows:

    /media
        /root: The instance's root volume.
        /ephemeral[0-N]: Instance store volumes.
        /persistent[0-N]: EBS volumes.
        /tmp: A temporary directory with lots of space.

WARNING: Be conscious about what this script prints to stdout, as that
         output is parsed by Flintrock.
"""
from __future__ import print_function
from __future__ import unicode_literals

import json
import platform
import subprocess
import sys

from collections import namedtuple

# Taken from: http://man7.org/linux/man-pages/man5/fstab.5.html
Mount = namedtuple(
    'Mount', [
        'device_name',
        'mount_point',
        'filesystem_type',
        'mount_options',
        'dump',
        'pass_number'
    ])

BlockDevice = namedtuple(
    'BlockDevice', sorted([
        'kname',
        'mountpoint',
        'size',
    ]))
BlockDevice.__new__.__defaults__ = (None, ) * len(BlockDevice._fields)


def device_pairs_to_tuple(pairs):
    device_dict = {}
    for pair in pairs:
        key, value = pair.split('=')
        key = key.lower()
        value = value.strip('"').lower()
        device_dict.update({key: value})
    return BlockDevice(**device_dict)


def get_non_root_block_devices():
    """
    Get all the non-root block devices available to the host.

    These are the devices we're going to format and mount for use.
    """
    block_devices_raw = subprocess.check_output([
        'lsblk',
        '--ascii',
        '--pairs',
        '--bytes',
        '--paths',
        '--output', 'KNAME,MOUNTPOINT,SIZE',
        # --inverse and --nodeps make sure that
        #   1) we get the mount points for devices that have holder devices
        #   2) we don't get the holder devices themselves
        '--inverse',
        '--nodeps',
        '--noheadings',
    ]).decode('utf-8')
    block_devices = [
        device_pairs_to_tuple(line.split())
        for line in block_devices_raw.splitlines()
    ]
    non_root_block_devices = [
        device for device in block_devices
        if device.mountpoint != '/'
    ]
    # Skip tiny devices, like the 1M devices that show up on
    # m5 instances on EC2.
    # See: https://github.com/nchammas/flintrock/issues/256
    non_trivial_non_root_block_devices = [
        device for device in non_root_block_devices
        if int(device.size) >= 1024 ** 3
    ]
    return non_trivial_non_root_block_devices


def unmount_devices(devices):
    """
    Unmount the provided devices.
    """
    with open('/proc/mounts') as m:
        mounts = [Mount(*line.split()) for line in m.read().splitlines()]

    for mount in mounts:
        if mount.device_name in [d.kname for d in devices]:
            subprocess.check_output(['sudo', 'umount', mount.device_name])


def format_devices(devices):
    """
    Create an ext4 filesystem on the provided devices.
    """
    format_processes = []
    for device in devices:
        p = subprocess.Popen([
            'sudo', 'mkfs.ext4',
            '-F',
            '-E',
            'lazy_itable_init=0,lazy_journal_init=0',
            device.kname],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        format_processes.append(p)

    for p in format_processes:
        stdout_raw, stderr_raw = p.communicate()
        stdout, stderr = stdout_raw.decode('utf-8'), stderr_raw.decode('utf-8')  # noqa
        return_code = p.returncode
        if return_code != 0:
            raise Exception(
                "Format process returned non-zero exit code: {code}\n{error}"
                .format(
                    code=return_code,
                    error=stderr))


def mount_devices(devices):
    """
    Mount the provided devices at the provided mount points.

    Additionally, add the appropriate entries to /etc/fstab so that the mounts
    persist across cluster stop/start.
    """
    for device in devices:
        subprocess.check_output([
            'sudo', 'mkdir', '-p', device.mountpoint])

        # Replace any existing fstab entries with our own.
        subprocess.check_output(
            """
            grep -v -e "^{device_name}" /etc/fstab | sudo tee /etc/fstab
            """.format(device_name=device.kname),
            shell=True)
        subprocess.check_output(
            """
            echo "{fstab_entry}" | sudo tee -a /etc/fstab
            """.format(fstab_entry='   '.join([
                device.kname,
                device.mountpoint,
                'ext4',
                'defaults,users,noatime',
                '0',
                '0'])),
            shell=True)

        subprocess.check_output([
            'sudo', 'mount', '--source', device.kname])
        # NOTE: `mount` changes the mount point owner to root, so we have
        #       to set it to what we want here, after `mount` runs.
        subprocess.check_output(
            'sudo chown "$(logname):$(logname)" {m}'.format(m=device.mountpoint),
            shell=True)


def create_root_dir():
    """
    Create a folder that services like HDFS and Spark can refer to to access
    local storage on the root volume.
    """
    path = '/media/root'
    subprocess.check_output([
        'sudo', 'mkdir', '-p', path])
    subprocess.check_output(
        'sudo chown "$(logname):$(logname)" {p}'.format(p=path),
        shell=True)
    return path


def create_tmp_dir(target):
    """
    Create a folder that services can use as a temporary directory for big files.
    """
    path = '/media/tmp'
    subprocess.check_output([
        'sudo', 'ln', '-s', target, path])
    subprocess.check_output(
        'sudo chown "$(logname):$(logname)" {p}'.format(p=path),
        shell=True)
    return path


if __name__ == '__main__':
    if sys.version_info < (2, 7) or ((3, 0) <= sys.version_info < (3, 4)):
        raise Exception(
            "This script is only supported on Python 2.7+ and 3.4+. "
            "You are running Python {v}.".format(v=platform.python_version()))

    non_root_block_devices = get_non_root_block_devices()

    # NOTE: For now we are assuming that all non-root devices are ephemeral devices.
    #       We're going to assign them the mount points we want them to have once we're
    #       done with the unmount -> format -> mount cycle.
    ephemeral_devices = []
    for (num, device) in enumerate(sorted(non_root_block_devices, key=lambda d: d.kname)):
        ephemeral_devices.append(
            BlockDevice(
                kname=device.kname,
                mountpoint='/media/ephemeral' + str(num)))

    unmount_devices(ephemeral_devices)
    format_devices(ephemeral_devices)
    mount_devices(ephemeral_devices)

    root_dir = create_root_dir()
    if ephemeral_devices:
        tmp_dir = ephemeral_devices[0].mountpoint
    else:
        tmp_dir = '/tmp'
    create_tmp_dir(tmp_dir)

    print(json.dumps(
        {
            'root': root_dir,
            'ephemeral': [d.mountpoint for d in ephemeral_devices]
        }))



import functools
import string
import sys
import time
import urllib.request
import base64
import logging
from collections import namedtuple
from datetime import datetime

# External modules
import boto3
import botocore
import click

# Flintrock modules
from .core import FlintrockCluster
from .core import provision_cluster
from .exceptions import (
    Error,
    ClusterNotFound,
    ClusterAlreadyExists,
    ClusterInvalidState,
    InterruptedEC2Operation,
    NothingToDo,
)
from .ssh import generate_ssh_key_pair


logger = logging.getLogger('flintrock.ec2')


class NoDefaultVPC(Error):
    def __init__(self, *, region: str):
        super().__init__(
            "Flintrock could not find a default VPC in {r}. "
            "Please explicitly specify a VPC to work with in that region. "
            "Flintrock does not support managing EC2 clusters outside a VPC."
            .format(r=region)
        )
        self.region = region


class ConfigurationNotSupported(Error):
    def __init__(self, message):
        super().__init__(message)


def timeit(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = datetime.now().replace(microsecond=0)
        res = func(*args, **kwargs)
        end = datetime.now().replace(microsecond=0)
        logger.info("{f} finished in {t}.".format(f=func.__name__, t=(end - start)))
        return res
    return wrapper


class EC2Cluster(FlintrockCluster):
    def __init__(
            self,
            region: str,
            vpc_id: str,
            master_instance: 'boto3.resources.factory.ec2.Instance',
            slave_instances: "List[boto3.resources.factory.ec2.Instance]",
            *args,
            **kwargs):
        super().__init__(*args, **kwargs)
        self.region = region
        self.vpc_id = vpc_id
        self.master_instance = master_instance
        self.slave_instances = slave_instances

    @property
    def instances(self):
        if self.master_instance:
            return [self.master_instance] + self.slave_instances
        else:
            return self.slave_instances

    @property
    def master_ip(self):
        return self.master_instance.public_ip_address

    @property
    def master_host(self):
        return self.master_instance.public_dns_name

    @property
    def slave_ips(self):
        return [i.public_ip_address for i in self.slave_instances]

    @property
    def slave_hosts(self):
        return [i.public_dns_name for i in self.slave_instances]

    @property
    def num_masters(self):
        return 1 if self.master_instance else 0

    @property
    def num_slaves(self):
        return len(self.slave_instances)

    @property
    def state(self):
        instance_states = set(
            instance.state['Name'] for instance in self.instances)
        if len(instance_states) == 1:
            return instance_states.pop()
        else:
            return 'inconsistent'

    def wait_for_state(self, state: str):
        """
        Wait for the cluster's instances to a reach a specific state.
        The state of any services installed on the cluster is a
        separate matter.

        This method updates the cluster's instance metadata and
        master and slave IP addresses and hostnames.
        """
        ec2 = boto3.resource(service_name='ec2', region_name=self.region)

        while any([i.state['Name'] != state for i in self.instances]):
            if logger.isEnabledFor(logging.DEBUG):
                waiting_instances = [i for i in self.instances if i.state['Name'] != state]
                sample = ', '.join(["'{}'".format(i.id) for i in waiting_instances][:3])
                logger.debug("{size} instances not in state '{state}': {sample}, ...".format(size=len(waiting_instances), state=state, sample=sample))
            time.sleep(3)
            # Update metadata for all instances in one shot. We don't want
            # to make a call to AWS for each of potentially hundreds of
            # instances.
            instances = list(
                ec2.instances.filter(
                    # NOTE: We use Filters instead of InstanceIds to avoid
                    #       the issue described here: https://github.com/boto/boto3/issues/479
                    Filters=[
                        {'Name': 'instance-id', 'Values': [i.id for i in self.instances]}
                    ]))
            (self.master_instance, self.slave_instances) = _get_cluster_master_slaves(instances)

    def destroy(self):
        self.destroy_check()
        super().destroy()
        ec2 = boto3.resource(service_name='ec2', region_name=self.region)

        # TODO: Centralize logic to get Flintrock base security group. (?)
        flintrock_base_group = list(
            ec2.security_groups.filter(
                Filters=[
                    {'Name': 'group-name', 'Values': ['flintrock']},
                    {'Name': 'vpc-id', 'Values': [self.vpc_id]},
                ]))[0]

        # We "unassign" the cluster security group here (i.e. the
        # 'flintrock-clustername' group) so that we can immediately delete it once
        # the instances are terminated. If we don't do this, we get dependency
        # violations for a couple of minutes before we can actually delete the group.
        # TODO: Is there a way to do this in one call for all instances?
        #       Do we need to throttle these calls?
        for instance in self.instances:
            instance.modify_attribute(
                Groups=[flintrock_base_group.id])

        # TODO: Centralize logic to get cluster security group name from cluster name.
        cluster_group = list(
            ec2.security_groups.filter(
                Filters=[
                    {'Name': 'group-name', 'Values': ['flintrock-' + self.name]},
                    {'Name': 'vpc-id', 'Values': [self.vpc_id]},
                ]))[0]
        cluster_group.delete()

        (ec2.instances
            .filter(
                Filters=[
                    {'Name': 'instance-id', 'Values': [i.id for i in self.instances]}
                ])
            .terminate())
        self.wait_for_state('terminated')

    def start_check(self):
        if self.state == 'running':
            raise NothingToDo("Cluster is already running.")
        elif self.state != 'stopped':
            raise ClusterInvalidState(
                attempted_command='start',
                state=self.state)

    @timeit
    def start(self, *, user: str, identity_file: str):
        # TODO: Do these _check() methods make sense here?
        self.start_check()
        ec2 = boto3.resource(service_name='ec2', region_name=self.region)
        (ec2.instances
            .filter(
                Filters=[
                    {'Name': 'instance-id', 'Values': [i.id for i in self.instances]}
                ])
            .start())
        self.wait_for_state('running')

        super().start(
            user=user,
            identity_file=identity_file)

    def stop_check(self):
        if self.state == 'stopped':
            raise NothingToDo("Cluster is already stopped.")
        elif self.state != 'running':
            raise ClusterInvalidState(
                attempted_command='stop',
                state=self.state)

    @timeit
    def stop(self):
        self.stop_check()
        super().stop()

        ec2 = boto3.resource(service_name='ec2', region_name=self.region)
        (ec2.instances
            .filter(
                Filters=[
                    {'Name': 'instance-id', 'Values': [i.id for i in self.instances]}
                ])
            .stop())
        self.wait_for_state('stopped')

    def add_slaves_check(self):
        if self.state != 'running':
            raise ClusterInvalidState(
                attempted_command='add-slaves',
                state=self.state)

    @timeit
    def add_slaves(
            self,
            *,
            user: str,
            identity_file: str,
            num_slaves: int,
            spot_price: float,
            min_root_ebs_size_gb: int,
            tags: list,
            assume_yes: bool):
        security_group_ids = [
            group['GroupId']
            for group in self.master_instance.security_groups]
        block_device_mappings = get_ec2_block_device_mappings(
            min_root_ebs_size_gb=min_root_ebs_size_gb,
            ami=self.master_instance.image_id,
            region=self.region)
        availability_zone = self.master_instance.placement['AvailabilityZone']

        ec2 = boto3.resource(service_name='ec2', region_name=self.region)
        client = ec2.meta.client

        response = client.describe_instance_attribute(
            InstanceId=self.master_instance.id,
            Attribute='instanceInitiatedShutdownBehavior'
        )
        instance_initiated_shutdown_behavior = response['InstanceInitiatedShutdownBehavior']['Value']

        response = client.describe_instance_attribute(
            InstanceId=self.master_instance.id,
            Attribute='userData'
        )
        if not response['UserData']:
            user_data = ''
        else:
            user_data = response['UserData']['Value']

        if not self.master_instance.iam_instance_profile:
            instance_profile_arn = ''
        else:
            instance_profile_arn = self.master_instance.iam_instance_profile['Arn']

        self.add_slaves_check()
        try:
            new_slave_instances = _create_instances(
                num_instances=num_slaves,
                region=self.region,
                spot_price=spot_price,
                ami=self.master_instance.image_id,
                assume_yes=assume_yes,
                key_name=self.master_instance.key_name,
                instance_type=self.master_instance.instance_type,
                block_device_mappings=block_device_mappings,
                availability_zone=availability_zone,
                placement_group=self.master_instance.placement['GroupName'],
                tenancy=self.master_instance.placement['Tenancy'],
                security_group_ids=security_group_ids,
                subnet_id=self.master_instance.subnet_id,
                instance_profile_arn=instance_profile_arn,
                ebs_optimized=self.master_instance.ebs_optimized,
                instance_initiated_shutdown_behavior=instance_initiated_shutdown_behavior,
                user_data=user_data)

            slave_tags = [
                {'Key': 'flintrock-role', 'Value': 'slave'},
                {'Key': 'Name', 'Value': '{c}-slave'.format(c=self.name)}]
            slave_tags += tags

            (ec2.instances
                .filter(
                    Filters=[
                        {'Name': 'instance-id', 'Values': [i.id for i in new_slave_instances]}
                    ])
                .create_tags(Tags=slave_tags))

            existing_slaves = {i.public_ip_address for i in self.slave_instances}

            self.slave_instances += new_slave_instances
            self.wait_for_state('running')

            new_slaves = {i.public_ip_address for i in self.slave_instances} - existing_slaves

            super().add_slaves(
                user=user,
                identity_file=identity_file,
                new_hosts=new_slaves)
        except (Exception, KeyboardInterrupt) as e:
            if isinstance(e, InterruptedEC2Operation):
                cleanup_instances = e.instances
            else:
                cleanup_instances = new_slave_instances
            _cleanup_instances(
                instances=cleanup_instances,
                assume_yes=assume_yes,
                region=self.region,
            )
            raise

    @timeit
    def remove_slaves(self, *, user: str, identity_file: str, num_slaves: int):
        ec2 = boto3.resource(service_name='ec2', region_name=self.region)

        # self.remove_slaves_check() (?)

        # Remove spot instances first, if any.
        _instances = sorted(
            self.slave_instances,
            key=lambda x: x.instance_lifecycle == 'spot',
            reverse=True)
        removed_slave_instances, self.slave_instances = \
            _instances[0:num_slaves], _instances[num_slaves:]

        if self.state == 'running':
            super().remove_slaves(user=user, identity_file=identity_file)

        # TODO: Centralize logic to get Flintrock base security group.
        flintrock_base_group = list(
            ec2.security_groups.filter(
                Filters=[
                    {'Name': 'group-name', 'Values': ['flintrock']},
                    {'Name': 'vpc-id', 'Values': [self.vpc_id]},
                ]))[0]

        # TODO: Is there a way to do this in one call for all instances?
        for instance in removed_slave_instances:
            instance.modify_attribute(
                Groups=[flintrock_base_group.id])

        (ec2.instances
            .filter(
                Filters=[
                    {'Name': 'instance-id', 'Values': [i.id for i in removed_slave_instances]}
                ])
            .terminate())

    def run_command_check(self):
        if self.state != 'running':
            raise ClusterInvalidState(
                attempted_command='run-command',
                state=self.state)

    @timeit
    def run_command(self, *, master_only, command, user, identity_file):
        self.run_command_check()
        super().run_command(
            master_only=master_only,
            user=user,
            identity_file=identity_file,
            command=command)

    def copy_file_check(self):
        if self.state != 'running':
            raise ClusterInvalidState(
                attempted_command='copy-file',
                state=self.state)

    @timeit
    def copy_file(self, *, local_path, remote_path, master_only=False, user, identity_file):
        self.copy_file_check()
        super().copy_file(
            master_only=master_only,
            user=user,
            identity_file=identity_file,
            local_path=local_path,
            remote_path=remote_path)

    def print(self):
        """
        Print information about the cluster to screen in YAML.

        We don't use PyYAML because we want to control the key order
        in the output.
        """
        # Mark the boundaries of the YAML output.
        # See: http://yaml.org/spec/current.html#id2525905
        # print('---')
        print(self.name + ':')
        print('  state: {s}'.format(s=self.state))
        print('  node-count: {nc}'.format(nc=len(self.instances)))
        if self.state == 'running':
            print('  master:', self.master_host if self.num_masters > 0 else '')
            print(
                '\n    - '.join(
                    ['  slaves:'] + (self.slave_hosts if self.num_slaves > 0 else [])))
        # print('...')


def get_default_vpc(region: str) -> 'boto3.resources.factory.ec2.Vpc':
    """
    Get the user's default VPC in the provided region.
    """
    ec2 = boto3.resource(service_name='ec2', region_name=region)

    default_vpc = list(
        ec2.vpcs.filter(
            Filters=[{'Name': 'isDefault', 'Values': ['true']}]))

    if default_vpc:
        return default_vpc[0]
    else:
        raise NoDefaultVPC(region=region)


def check_network_config(*, region_name: str, vpc_id: str, subnet_id: str):
    """
    Check that the VPC and subnet are configured to allow Flintrock to create
    clusters.

    Currently, Flintrock requires DNS names and public IPs to be enabled.
    """
    ec2 = boto3.resource(service_name='ec2', region_name=region_name)

    if not ec2.Vpc(vpc_id).describe_attribute(Attribute='enableDnsHostnames')['EnableDnsHostnames']['Value']:
        raise ConfigurationNotSupported(
            "{v} does not have DNS hostnames enabled. "
            "Flintrock requires DNS hostnames to be enabled.\n"
            "See: https://github.com/nchammas/flintrock/issues/43"
            .format(v=vpc_id)
        )
    if not ec2.Subnet(subnet_id).map_public_ip_on_launch:
        raise ConfigurationNotSupported(
            "{s} does not auto-assign public IP addresses. "
            "Flintrock requires public IP addresses.\n"
            "See: https://github.com/nchammas/flintrock/issues/14"
            .format(s=subnet_id)
        )


def get_security_groups(
        *,
        vpc_id,
        region,
        security_group_names) -> "List[boto3.resource('ec2').SecurityGroup]":
    ec2 = boto3.resource(service_name='ec2', region_name=region)

    groups = list(
        ec2.security_groups.filter(
            Filters=[
                {'Name': 'group-name', 'Values': security_group_names},
                {'Name': 'vpc-id', 'Values': [vpc_id]},
            ]))

    found_group_names = [group.group_name for group in groups]
    missing_group_names = set(security_group_names) - set(found_group_names)
    if missing_group_names:
        raise Error(
            "Could not find the following security group{s}: {groups}"
            .format(
                s='' if len(missing_group_names) == 1 else 's',
                groups=', '.join(list(missing_group_names))))

    return groups


def get_or_create_flintrock_security_groups(
        *,
        cluster_name,
        vpc_id,
        region) -> "List[boto3.resource('ec2').SecurityGroup]":
    """
    If they do not already exist, create all the security groups needed for a
    Flintrock cluster.
    """
    ec2 = boto3.resource(service_name='ec2', region_name=region)

    SecurityGroupRule = namedtuple(
        'SecurityGroupRule', [
            'ip_protocol',
            'from_port',
            'to_port',
            'src_group',
            'cidr_ip'])

    # TODO: Make these into methods, since we need this logic (though simple)
    #       in multiple places. (?)
    flintrock_group_name = 'flintrock'
    cluster_group_name = 'flintrock-' + cluster_name

    # The Flintrock group is common to all Flintrock clusters and authorizes client traffic
    # to them.
    flintrock_group = list(
        ec2.security_groups.filter(
            Filters=[
                {'Name': 'group-name', 'Values': [flintrock_group_name]},
                {'Name': 'vpc-id', 'Values': [vpc_id]},
            ]))
    flintrock_group = flintrock_group[0] if flintrock_group else None

    # The cluster group is specific to one Flintrock cluster and authorizes intra-cluster
    # communication.
    cluster_group = list(
        ec2.security_groups.filter(
            Filters=[
                {'Name': 'group-name', 'Values': [cluster_group_name]},
                {'Name': 'vpc-id', 'Values': [vpc_id]},
            ]))
    cluster_group = cluster_group[0] if cluster_group else None

    if not flintrock_group:
        flintrock_group = ec2.create_security_group(
            GroupName=flintrock_group_name,
            Description="Flintrock base group",
            VpcId=vpc_id)

    # Rules for the client interacting with the cluster.
    flintrock_client_ip = (
        urllib.request.urlopen('https://checkip.amazonaws.com/')
        .read().decode('utf-8').strip())
    flintrock_client_cidr = '{ip}/32'.format(ip=flintrock_client_ip)

    # TODO: Services should be responsible for registering what ports they want exposed.
    client_rules = [
        # SSH
        SecurityGroupRule(
            ip_protocol='tcp',
            from_port=22,
            to_port=22,
            cidr_ip=flintrock_client_cidr,
            src_group=None),
        # HDFS
        SecurityGroupRule(
            ip_protocol='tcp',
            from_port=50070,
            to_port=50070,
            cidr_ip=flintrock_client_cidr,
            src_group=None),
        # Spark
        SecurityGroupRule(
            ip_protocol='tcp',
            from_port=8080,
            to_port=8081,
            cidr_ip=flintrock_client_cidr,
            src_group=None),
        SecurityGroupRule(
            ip_protocol='tcp',
            from_port=4040,
            to_port=4050,
            cidr_ip=flintrock_client_cidr,
            src_group=None),
        SecurityGroupRule(
            ip_protocol='tcp',
            from_port=7077,
            to_port=7077,
            cidr_ip=flintrock_client_cidr,
            src_group=None),
        # Spark REST Server
        SecurityGroupRule(
            ip_protocol='tcp',
            from_port=6066,
            to_port=6066,
            cidr_ip=flintrock_client_cidr,
            src_group=None)
    ]

    # TODO: Don't try adding rules that already exist.
    # TODO: Add rules in one shot.
    for rule in client_rules:
        try:
            flintrock_group.authorize_ingress(
                IpProtocol=rule.ip_protocol,
                FromPort=rule.from_port,
                ToPort=rule.to_port,
                CidrIp=rule.cidr_ip)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] != 'InvalidPermission.Duplicate':
                raise Exception("Error adding rule: {r}".format(r=rule))

    # Rules for internal cluster communication.
    if not cluster_group:
        cluster_group = ec2.create_security_group(
            GroupName=cluster_group_name,
            Description="Flintrock cluster group",
            VpcId=vpc_id)

    try:
        cluster_group.authorize_ingress(
            IpPermissions=[
                {
                    'IpProtocol': '-1',  # -1 means all
                    'FromPort': -1,
                    'ToPort': -1,
                    'UserIdGroupPairs': [{'GroupId': cluster_group.id}]
                }])
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] != 'InvalidPermission.Duplicate':
            raise Exception("Error authorizing cluster ingress to self.") from e

    return [flintrock_group, cluster_group]


def get_ec2_block_device_mappings(
        *,
        min_root_ebs_size_gb: int,
        ami: str,
        region: str) -> 'List[dict]':
    """
    Get the block device map we should assign to instances launched from a given AMI.

    This is how we configure storage on the instance.
    """
    ec2 = boto3.resource(service_name='ec2', region_name=region)
    block_device_mappings = []

    try:
        image = list(
            ec2.images.filter(
                Filters=[
                    {'Name': 'image-id', 'Values': [ami]}
                ]))[0]
    except IndexError as e:
        raise Error(
            "Error: Could not find {ami} in region {region}.".format(
                ami=ami,
                region=region))

    if image.root_device_type == 'ebs':
        root_device = [
            device for device in image.block_device_mappings
            if device['DeviceName'] == image.root_device_name][0]
        if root_device['Ebs']['VolumeSize'] < min_root_ebs_size_gb:
            root_device['Ebs'].update({
                # Max root volume size for instance store-backed AMIs is 10 GiB.
                # See: http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/add-instance-store-volumes.html
                # Though, this code is probably incorrect for instance store-backed
                # instances anyway, since boto3 doesn't seem to let you set the size
                # of a root instance store volume.
                'VolumeSize': min_root_ebs_size_gb,
                # gp2 is general-purpose SSD
                'VolumeType': 'gp2'})
        del root_device['Ebs']['Encrypted']
        block_device_mappings.append(root_device)

    for i in range(12):
        ephemeral_device = {
            'VirtualName': 'ephemeral' + str(i),
            'DeviceName': '/dev/sd' + string.ascii_lowercase[i + 1]}
        block_device_mappings.append(ephemeral_device)

    return block_device_mappings


def _create_instances(
        *,
        num_instances,
        region,
        spot_price,
        ami,
        assume_yes,
        key_name,
        instance_type,
        block_device_mappings,
        availability_zone,
        placement_group,
        tenancy,
        security_group_ids,
        subnet_id,
        instance_profile_arn,
        ebs_optimized,
        instance_initiated_shutdown_behavior,
        user_data) -> 'List[boto3.resources.factory.ec2.Instance]':
    ec2 = boto3.resource(service_name='ec2', region_name=region)

    cluster_instances = []
    spot_requests = []

    try:
        if spot_price:
            user_data = base64.b64encode(user_data.encode('utf-8')).decode()
            logger.info("Requesting {c} spot instances at a max price of ${p}...".format(
                c=num_instances, p=spot_price))
            client = ec2.meta.client
            spot_requests = client.request_spot_instances(
                SpotPrice=str(spot_price),
                InstanceCount=num_instances,
                LaunchSpecification={
                    'ImageId': ami,
                    'KeyName': key_name,
                    'InstanceType': instance_type,
                    'BlockDeviceMappings': block_device_mappings,
                    'Placement': {
                        'AvailabilityZone': availability_zone,
                        'GroupName': placement_group},
                    'SecurityGroupIds': security_group_ids,
                    'SubnetId': subnet_id,
                    'IamInstanceProfile': {
                        'Arn': instance_profile_arn},
                    'EbsOptimized': ebs_optimized,
                    'UserData': user_data})['SpotInstanceRequests']

            request_ids = [r['SpotInstanceRequestId'] for r in spot_requests]
            pending_request_ids = request_ids

            while pending_request_ids:
                logger.info("{grant} of {req} instances granted. Waiting...".format(
                    grant=num_instances - len(pending_request_ids),
                    req=num_instances))
                time.sleep(30)
                spot_requests = client.describe_spot_instance_requests(
                    SpotInstanceRequestIds=request_ids)['SpotInstanceRequests']

                failed_requests = [r for r in spot_requests if r['State'] == 'failed']
                if failed_requests:
                    failure_reasons = {r['Status']['Code'] for r in failed_requests}
                    raise Error(
                        "The spot request failed for the following reason{s}: {reasons}"
                        .format(
                            s='' if len(failure_reasons) == 1 else 's',
                            reasons=', '.join(failure_reasons)))

                pending_request_ids = [
                    r['SpotInstanceRequestId'] for r in spot_requests
                    if r['State'] == 'open']

            logger.info("All {c} instances granted.".format(c=num_instances))

            cluster_instances = list(
                ec2.instances.filter(
                    Filters=[
                        {'Name': 'instance-id', 'Values': [r['InstanceId'] for r in spot_requests]}
                    ]))
        else:
            # Move this to flintrock.py?
            logger.info("Launching {c} instance{s}...".format(
                c=num_instances,
                s='' if num_instances == 1 else 's'))

            # TODO: If an exception is raised in here, some instances may be
            #       left stranded.
            cluster_instances = ec2.create_instances(
                MinCount=num_instances,
                MaxCount=num_instances,
                ImageId=ami,
                KeyName=key_name,
                InstanceType=instance_type,
                BlockDeviceMappings=block_device_mappings,
                Placement={
                    'AvailabilityZone': availability_zone,
                    'Tenancy': tenancy,
                    'GroupName': placement_group},
                SecurityGroupIds=security_group_ids,
                SubnetId=subnet_id,
                IamInstanceProfile={
                    'Arn': instance_profile_arn},
                EbsOptimized=ebs_optimized,
                InstanceInitiatedShutdownBehavior=instance_initiated_shutdown_behavior,
                UserData=user_data)
        time.sleep(10)  # AWS metadata eventual consistency tax.
        return cluster_instances
    except (Exception, KeyboardInterrupt) as e:
        if not isinstance(e, KeyboardInterrupt):
            print(e, file=sys.stderr)
        if spot_requests:
            request_ids = [r['SpotInstanceRequestId'] for r in spot_requests]
            if any([r['State'] != 'active' for r in spot_requests]):
                print("Canceling spot instance requests...", file=sys.stderr)
                client.cancel_spot_instance_requests(
                    SpotInstanceRequestIds=request_ids)
            # Make sure we have the latest information on any launched spot instances.
            spot_requests = client.describe_spot_instance_requests(
                SpotInstanceRequestIds=request_ids)['SpotInstanceRequests']
            instance_ids = [
                r['InstanceId'] for r in spot_requests
                if 'InstanceId' in r]
            if instance_ids:
                cluster_instances = list(
                    ec2.instances.filter(
                        Filters=[
                            {'Name': 'instance-id', 'Values': instance_ids}
                        ]))
        raise InterruptedEC2Operation(instances=cluster_instances) from e


@timeit
def launch(
        *,
        cluster_name,
        num_slaves,
        services,
        assume_yes,
        key_name,
        identity_file,
        instance_type,
        region,
        availability_zone,
        ami,
        user,
        security_groups,
        spot_price=None,
        min_root_ebs_size_gb,
        vpc_id,
        subnet_id,
        instance_profile_name,
        placement_group,
        tenancy='default',
        ebs_optimized=False,
        instance_initiated_shutdown_behavior='stop',
        user_data,
        tags):
    """
    Launch a cluster.
    """
    if not vpc_id:
        vpc_id = get_default_vpc(region=region).id
    else:
        # If it's a non-default VPC -- i.e. the user set it up -- make sure it's
        # configured correctly.
        check_network_config(
            region_name=region,
            vpc_id=vpc_id,
            subnet_id=subnet_id)

    try:
        get_cluster(
            cluster_name=cluster_name,
            region=region,
            vpc_id=vpc_id)
    except ClusterNotFound as e:
        pass
    else:
        raise ClusterAlreadyExists(
            "Cluster {c} already exists in region {r}, VPC {v}.".format(
                c=cluster_name,
                r=region,
                v=vpc_id))

    flintrock_security_groups = get_or_create_flintrock_security_groups(
        cluster_name=cluster_name,
        vpc_id=vpc_id,
        region=region)
    user_security_groups = get_security_groups(
        vpc_id=vpc_id,
        region=region,
        security_group_names=security_groups)
    security_group_ids = [sg.id for sg in user_security_groups + flintrock_security_groups]
    block_device_mappings = get_ec2_block_device_mappings(
        min_root_ebs_size_gb=min_root_ebs_size_gb,
        ami=ami,
        region=region)

    ec2 = boto3.resource(service_name='ec2', region_name=region)
    iam = boto3.resource(service_name='iam', region_name=region)

    # We use IAM profile ARNs internally because AWS's API prefers that in
    # a few places.
    # See: https://github.com/boto/boto3/issues/769
    if instance_profile_name:
        instance_profile_arn = iam.InstanceProfile(instance_profile_name).arn
    else:
        instance_profile_arn = ''

    num_instances = num_slaves + 1
    if user_data is not None:
        user_data = user_data.read()
    else:
        user_data = ''

    try:
        cluster_instances = _create_instances(
            num_instances=num_instances,
            region=region,
            spot_price=spot_price,
            ami=ami,
            assume_yes=assume_yes,
            key_name=key_name,
            instance_type=instance_type,
            block_device_mappings=block_device_mappings,
            availability_zone=availability_zone,
            placement_group=placement_group,
            tenancy=tenancy,
            security_group_ids=security_group_ids,
            subnet_id=subnet_id,
            instance_profile_arn=instance_profile_arn,
            ebs_optimized=ebs_optimized,
            instance_initiated_shutdown_behavior=instance_initiated_shutdown_behavior,
            user_data=user_data)

        master_instance = cluster_instances[0]
        slave_instances = cluster_instances[1:]

        master_tags = [
            {'Key': 'flintrock-role', 'Value': 'master'},
            {'Key': 'Name', 'Value': '{c}-master'.format(c=cluster_name)}]
        master_tags += tags

        (ec2.instances
            .filter(
                Filters=[
                    {'Name': 'instance-id', 'Values': [master_instance.id]}
                ])
            .create_tags(Tags=master_tags))

        slave_tags = [
            {'Key': 'flintrock-role', 'Value': 'slave'},
            {'Key': 'Name', 'Value': '{c}-slave'.format(c=cluster_name)}]
        slave_tags += tags

        (ec2.instances
            .filter(
                Filters=[
                    {'Name': 'instance-id', 'Values': [i.id for i in slave_instances]}
                ])
            .create_tags(Tags=slave_tags))

        cluster = EC2Cluster(
            name=cluster_name,
            region=region,
            vpc_id=vpc_id,
            ssh_key_pair=generate_ssh_key_pair(),
            master_instance=master_instance,
            slave_instances=slave_instances)

        cluster.wait_for_state('running')

        provision_cluster(
            cluster=cluster,
            services=services,
            user=user,
            identity_file=identity_file)

        return cluster
    except (Exception, KeyboardInterrupt) as e:
        if isinstance(e, InterruptedEC2Operation):
            cleanup_instances = e.instances
        else:
            # TODO: There is no guarantee that cluster_instances is
            #       defined.
            # See: https://github.com/nchammas/flintrock/issues/183
            cleanup_instances = cluster_instances
        _cleanup_instances(
            instances=cleanup_instances,
            assume_yes=assume_yes,
            region=region,
        )
        raise


def get_cluster(*, cluster_name: str, region: str, vpc_id: str) -> EC2Cluster:
    """
    Get an existing EC2 cluster.
    """
    cluster = get_clusters(
        cluster_names=[cluster_name],
        region=region,
        vpc_id=vpc_id)
    return cluster[0]


def get_clusters(*, cluster_names: list=[], region: str, vpc_id: str) -> list:
    """
    Get all the named clusters. If no names are given, get all clusters.

    We do a little extra work here so that we only make one call to AWS
    regardless of how many clusters we have to look up. That's because querying
    AWS -- a network operation -- is by far the slowest step.
    """
    ec2 = boto3.resource(service_name='ec2', region_name=region)
    if not vpc_id:
        vpc_id = get_default_vpc(region=region).id

    if cluster_names:
        group_name_filter = ['flintrock-' + cn for cn in cluster_names]
    else:
        group_name_filter = ['flintrock']

    all_clusters_instances = list(
        ec2.instances.filter(
            Filters=[
                {'Name': 'instance-state-name', 'Values': ['pending', 'running', 'stopping', 'stopped']},
                {'Name': 'instance.group-name', 'Values': group_name_filter},
                {'Name': 'vpc-id', 'Values': [vpc_id]},
            ]))

    found_cluster_names = {
        _get_cluster_name(instance) for instance in all_clusters_instances}

    if cluster_names:
        missing_cluster_names = set(cluster_names) - found_cluster_names
        if missing_cluster_names:
            raise ClusterNotFound("No cluster {c} in region {r}.".format(
                c=missing_cluster_names.pop(),
                r=region))

    clusters = [
        _compose_cluster(
            name=cluster_name,
            region=region,
            vpc_id=vpc_id,
            instances=list(filter(
                lambda x: _get_cluster_name(x) == cluster_name, all_clusters_instances)))
        for cluster_name in found_cluster_names]

    return clusters


def cli_validate_tags(ctx, param, value):
    return validate_tags(value)


def validate_tags(value):
    """
    Validate and parse optional EC2 tags.
    """
    err_msg = ("Tags need to be specified as 'Key,Value' pairs "
               "separated by a single comma. Key cannot be empty "
               "or be made up entirely of whitespace.")
    tags = value
    result = []
    for tag in tags:
        if tag.count(',') != 1:
            raise click.BadParameter(err_msg)
        key, value = [word.strip() for word in tag.split(',', maxsplit=1)]
        if not key:
            raise click.BadParameter(err_msg)
        result.append({'Key': key, 'Value': value})

    return result


def _get_cluster_name(instance: 'boto3.resources.factory.ec2.Instance') -> str:
    """
    Given an EC2 instance, get the name of the Flintrock cluster it belongs to.
    """
    for group in instance.security_groups:
        if group['GroupName'].startswith('flintrock-'):
            return group['GroupName'].replace('flintrock-', '', 1)
    else:
        raise Exception("Could not extract cluster name from instance: {i}".format(
            i=instance.id))


def _get_cluster_master_slaves(
        instances: list) -> ('boto3.resources.factory.ec2.Instance', list):
    """
    Get the master and slave instances from a set of raw EC2 instances representing
    a Flintrock cluster.
    """
    master_instance = None
    slave_instances = []

    for instance in instances:
        if not instance.tags:
            # TODO: Better handle malformed clusters with missing tags.
            # See: https://github.com/nchammas/flintrock/issues/183
            continue
        for tag in instance.tags:
            if tag['Key'] == 'flintrock-role':
                if tag['Value'] == 'master':
                    if master_instance is not None:
                        raise Exception("More than one master found.")
                    else:
                        master_instance = instance
                        break
                elif tag['Value'] == 'slave':
                    slave_instances.append(instance)

    # if not master_instance:
    #     print("Warning: No master found.", file=sys.stderr)
    # elif not slave_instances:
    #     print("Warning: No slaves found.", file=sys.stderr)

    return (master_instance, slave_instances)


def _compose_cluster(*, name: str, region: str, vpc_id: str, instances: list) -> EC2Cluster:
    """
    Compose an EC2Cluster object from a set of raw EC2 instances representing
    a Flintrock cluster.
    """
    (master_instance, slave_instances) = _get_cluster_master_slaves(instances)

    cluster = EC2Cluster(
        name=name,
        region=region,
        vpc_id=vpc_id,
        master_instance=master_instance,
        slave_instances=slave_instances)

    return cluster


def _cleanup_instances(*, instances: list, assume_yes: bool, region: str):
    ec2 = boto3.resource(service_name='ec2', region_name=region)
    if instances:
        if not assume_yes:
            yes = click.confirm(
                text="Do you want to terminate the {c} instances created by this operation?"
                     .format(c=len(instances)),
                err=True,
                default=True)

        if assume_yes or yes:
            print("Terminating instances...", file=sys.stderr)
            (ec2.instances
                .filter(
                    Filters=[
                        {'Name': 'instance-id', 'Values': [i.id for i in instances]}
                    ])
                .terminate())



import setuptools
from flintrock import __version__


with open('README.md') as f:
    long_description = f.read()

setuptools.setup(
    name='Flintrock',
    version=__version__,
    description='A command-line tool for launching Apache Spark clusters.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/nchammas/flintrock',
    author='Nicholas Chammas',
    author_email='nicholas.chammas@gmail.com',
    license='Apache License 2.0',
    python_requires='>= 3.4',

    # See: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        'Topic :: Utilities',
        'Environment :: Console',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords=['Apache Spark'],

    packages=setuptools.find_packages(),
    include_package_data=True,

    # We pin dependencies because sometimes projects do not
    # strictly follow semantic versioning, so new "feature"
    # releases end up making backwards-incompatible changes.
    # Sometimes, new releases even introduce bugs which
    # totally break Flintrock.
    # For example: https://github.com/paramiko/paramiko/issues/615
    install_requires=[
        'boto3 == 1.9.57',
        'botocore == 1.12.57',
        'click == 7.0',
        'paramiko == 2.4.2',
        'PyYAML == 3.13',
        # This is to address reports that PyInstaller-packaged versions
        # of Flintrock intermittently fail due to an out-of-date version
        # of Cryptography being used.
        # See: https://github.com/nchammas/flintrock/issues/169
        'cryptography >= 1.7.2',
    ],

    entry_points={
        'console_scripts': [
            'flintrock = flintrock.__main__:main',
        ],
    },
)



from __future__ import print_function

import argparse
import errno
import os.path
import sys
import subprocess
import time

MAX_TRIES = 5


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('url')
    parser.add_argument('destination_dir')
    args = parser.parse_args()
    return (args.url, args.destination_dir)


if __name__ == '__main__':
    url, destination_dir = parse_args()

    try:
        os.makedirs(destination_dir, mode=0o755)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise

    download_path = '{}.download'.format(os.path.basename(destination_dir))

    tries = 0
    while True:
        try:
            subprocess.check_call(['curl', '--location', '--output', download_path, url])
            subprocess.check_call(['gzip', '--test', download_path])
            subprocess.check_call(['tar', 'xzf', download_path, '-C', destination_dir, '--strip-components=1'])
            subprocess.check_call(['rm', download_path])
        except subprocess.CalledProcessError as e:
            print(e, file=sys.stderr)
            if tries < MAX_TRIES:
                tries += 1
                time.sleep(1)
            else:
                print(
                    "Failed to download and unpack '{url}' after {tries} tries."
                    .format(
                        url=url,
                        tries=MAX_TRIES,
                    ),
                    file=sys.stderr,
                )
                sys.exit(1)
        else:
            break



import json
import os
import shlex
import socket
import sys
import urllib.error
import urllib.request
import logging

# External modules
import paramiko

# Flintrock modules
from .core import (
    FlintrockCluster,
    generate_template_mapping,
    get_formatted_template,
)
from .ssh import ssh_check_output

FROZEN = getattr(sys, 'frozen', False)

if FROZEN:
    THIS_DIR = sys._MEIPASS
else:
    THIS_DIR = os.path.dirname(os.path.realpath(__file__))

SCRIPTS_DIR = os.path.join(THIS_DIR, 'scripts')


logger = logging.getLogger('flintrock.services')


class FlintrockService:
    """
    This is an abstract class. Implementations of this class capture all the logic
    required to fully install and manage services like Spark on Flintrock clusters.
    """

    def __init__(self):
        """
        This is the only method signature that implementations don't have to follow.
        Use this method to set properties like the service version or download source
        which the rest of the methods here will need to do their work.
        """
        raise NotImplementedError

    def install(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        """
        Install the service on a node via the provided SSH client. This typically
        means downloading a software package and maybe even building it if necessary.

        This method is role-agnostic; it runs on both the cluster master and slaves.
        This method is meant to be called asynchronously.
        """
        raise NotImplementedError

    def configure(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        """
        Configure the installed service on a node via the provided SSH client. This
        typically means using templates to create configuration files on the node.

        This method is role-agnostic; it runs on both the cluster master and slaves.
        This method is meant to be called asynchronously.
        """
        raise NotImplementedError

    def configure_master(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        """
        Configure the service master on a node via the provided SSH client after the
        role-agnostic configuration in configure() is complete. Start the master and
        slaves.

        This method is meant to be called once on the cluster master.
        This method is meant to be called asynchronously.
        """
        raise NotImplementedError

    def configure_slave(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        """
        Configure a service slave on a node via the provided SSH client after the
        role-agnostic configuration in configure() is complete.

        This method is meant to be called once on each cluster slave.
        This method is meant to be called asynchronously.
        """
        raise NotImplementedError

    def health_check(
            self,
            master_host: str):
        """
        Check that the service is up and running by querying the cluster master.
        """
        raise NotImplementedError


class HDFS(FlintrockService):
    def __init__(self, *, version, download_source):
        self.version = version
        self.download_source = download_source
        self.manifest = {'version': version, 'download_source': download_source}

    def install(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        logger.info("[{h}] Installing HDFS...".format(
            h=ssh_client.get_transport().getpeername()[0]))

        with ssh_client.open_sftp() as sftp:
            sftp.put(
                localpath=os.path.join(SCRIPTS_DIR, 'download-package.py'),
                remotepath='/tmp/download-package.py')

        ssh_check_output(
            client=ssh_client,
            command="""
                set -e

                python /tmp/download-package.py "{download_source}" "hadoop"

                for f in $(find hadoop/bin -type f -executable -not -name '*.cmd'); do
                    sudo ln -s "$(pwd)/$f" "/usr/local/bin/$(basename $f)"
                done

                echo "export HADOOP_LIBEXEC_DIR='$(pwd)/hadoop/libexec'" >> .bashrc
            """.format(
                version=self.version,
                download_source=self.download_source.format(v=self.version),
            ))

    def configure(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        # TODO: os.walk() through these files.
        template_paths = [
            'hadoop/conf/masters',
            'hadoop/conf/slaves',
            'hadoop/conf/hadoop-env.sh',
            'hadoop/conf/core-site.xml',
            'hadoop/conf/hdfs-site.xml',
        ]

        ssh_check_output(
            client=ssh_client,
            command="mkdir -p hadoop/conf",
        )

        for template_path in template_paths:
            ssh_check_output(
                client=ssh_client,
                command="""
                    echo {f} > {p}
                """.format(
                    f=shlex.quote(
                        get_formatted_template(
                            path=os.path.join(THIS_DIR, "templates", template_path),
                            mapping=generate_template_mapping(
                                cluster=cluster,
                                hadoop_version=self.version,
                                # Hadoop doesn't need to know what
                                # Spark version we're using.
                                spark_version='',
                                spark_executor_instances=0,
                            ))),
                    p=shlex.quote(template_path)))

    # TODO: Convert this into start_master() and split master- or slave-specific
    #       stuff out of configure() into configure_master() and configure_slave().
    def configure_master(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        host = ssh_client.get_transport().getpeername()[0]
        logger.info("[{h}] Configuring HDFS master...".format(h=host))

        ssh_check_output(
            client=ssh_client,
            command="""
                # `|| true` because on cluster restart this command will fail.
                ./hadoop/bin/hdfs namenode -format -nonInteractive || true
            """)

        # This loop is a band-aid for: https://github.com/nchammas/flintrock/issues/157
        attempt_limit = 3
        for attempt in range(attempt_limit):
            try:
                ssh_check_output(
                    client=ssh_client,
                    command="""
                        ./hadoop/sbin/stop-dfs.sh
                        ./hadoop/sbin/start-dfs.sh

                        master_ui_response_code=0
                        while [ "$master_ui_response_code" -ne 200 ]; do
                            sleep 1
                            master_ui_response_code="$(
                                curl --head --silent --output /dev/null \
                                    --write-out "%{{http_code}}" {m}:50070
                            )"
                        done
                    """.format(m=shlex.quote(cluster.master_host)),
                    timeout_seconds=90
                )
                break
            except socket.timeout as e:
                logger.debug(
                    "Timed out waiting for HDFS master to come up.{}"
                    .format(" Trying again..." if attempt < attempt_limit - 1 else "")
                )
        else:
            raise Exception("Time out waiting for HDFS master to come up.")

    def health_check(self, master_host: str):
        # This info is not helpful as a detailed health check, but it gives us
        # an up / not up signal.
        hdfs_master_ui = 'http://{m}:50070/webhdfs/v1/?op=GETCONTENTSUMMARY'.format(m=master_host)

        try:
            json.loads(
                urllib.request
                .urlopen(hdfs_master_ui)
                .read()
                .decode('utf-8'))
            logger.info("HDFS online.")
        except Exception as e:
            raise Exception("HDFS health check failed.") from e


class Spark(FlintrockService):
    def __init__(
        self,
        *,
        spark_executor_instances: int,
        version: str=None,
        hadoop_version: str,
        download_source: str=None,
        git_commit: str=None,
        git_repository: str=None
    ):
        # TODO: Convert these checks into something that throws a proper exception.
        #       Perhaps reuse logic from CLI.
        assert bool(version) ^ bool(git_commit)
        if git_commit:
            assert git_repository

        self.spark_executor_instances = spark_executor_instances
        self.version = version
        self.hadoop_version = hadoop_version
        self.download_source = download_source
        self.git_commit = git_commit
        self.git_repository = git_repository

        self.manifest = {
            'version': version,
            'spark_executor_instances': spark_executor_instances,
            'hadoop_version': hadoop_version,
            'download_source': download_source,
            'git_commit': git_commit,
            'git_repository': git_repository}

    def install(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        logger.info("[{h}] Installing Spark...".format(
            h=ssh_client.get_transport().getpeername()[0]))

        if self.version:
            with ssh_client.open_sftp() as sftp:
                sftp.put(
                    localpath=os.path.join(SCRIPTS_DIR, 'download-package.py'),
                    remotepath='/tmp/download-package.py')

            ssh_check_output(
                client=ssh_client,
                command="""
                    python /tmp/download-package.py "{download_source}" "spark"
                """.format(
                    version=self.version,
                    download_source=self.download_source.format(v=self.version),
                ))

        else:
            ssh_check_output(
                client=ssh_client,
                command="""
                    set -e
                    sudo yum install -y git
                    sudo yum install -y java-devel
                    """)
            ssh_check_output(
                client=ssh_client,
                command="""
                    set -e
                    git clone {repo} spark
                    cd spark
                    git reset --hard {commit}
                    if [ -e "make-distribution.sh" ]; then
                        ./make-distribution.sh -Phadoop-{hadoop_short_version}
                    else
                        ./dev/make-distribution.sh -Phadoop-{hadoop_short_version}
                    fi
                """.format(
                    repo=shlex.quote(self.git_repository),
                    commit=shlex.quote(self.git_commit),
                    # Hardcoding this here until we figure out a better way to handle
                    # the supported build profiles.
                    hadoop_short_version='2.7',
                ))
        ssh_check_output(
            client=ssh_client,
            command="""
                set -e
                for f in $(find spark/bin -type f -executable -not -name '*.cmd'); do
                    sudo ln -s "$(pwd)/$f" "/usr/local/bin/$(basename $f)"
                done
                echo "export SPARK_HOME='$(pwd)/spark'" >> .bashrc
            """)

    def configure(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):

        template_paths = [
            'spark/conf/spark-env.sh',
            'spark/conf/slaves',
        ]

        ssh_check_output(
            client=ssh_client,
            command="mkdir -p spark/conf",
        )

        for template_path in template_paths:
            ssh_check_output(
                client=ssh_client,
                command="""
                    echo {f} > {p}
                """.format(
                    f=shlex.quote(
                        get_formatted_template(
                            path=os.path.join(THIS_DIR, "templates", template_path),
                            mapping=generate_template_mapping(
                                cluster=cluster,
                                spark_executor_instances=self.spark_executor_instances,
                                hadoop_version=self.hadoop_version,
                                spark_version=self.version or self.git_commit,
                            ))),
                    p=shlex.quote(template_path)))

    # TODO: Convert this into start_master() and split master- or slave-specific
    #       stuff out of configure() into configure_master() and configure_slave().
    #       start_slave() can block until slave is fully up; that way we don't need
    #       a sleep() before starting the master.
    def configure_master(
            self,
            ssh_client: paramiko.client.SSHClient,
            cluster: FlintrockCluster):
        host = ssh_client.get_transport().getpeername()[0]
        logger.info("[{h}] Configuring Spark master...".format(h=host))

        # This loop is a band-aid for: https://github.com/nchammas/flintrock/issues/129
        attempt_limit = 3
        for attempt in range(attempt_limit):
            try:
                ssh_check_output(
                    client=ssh_client,
                    # Maybe move this shell script out to some separate
                    # file/folder for the Spark service.
                    command="""
                        spark/sbin/start-all.sh

                        master_ui_response_code=0
                        while [ "$master_ui_response_code" -ne 200 ]; do
                            sleep 1
                            master_ui_response_code="$(
                                curl --head --silent --output /dev/null \
                                    --write-out "%{{http_code}}" {m}:8080
                            )"
                        done
                    """.format(m=shlex.quote(cluster.master_host)),
                    timeout_seconds=90
                )
                break
            except socket.timeout as e:
                logger.debug(
                    "Timed out waiting for Spark master to come up.{}"
                    .format(" Trying again..." if attempt < attempt_limit - 1 else "")
                )
        else:
            raise Exception("Timed out waiting for Spark master to come up.")

    def health_check(self, master_host: str):
        spark_master_ui = 'http://{m}:8080/json/'.format(m=master_host)

        try:
            json.loads(
                urllib.request
                .urlopen(spark_master_ui)
                .read()
                .decode('utf-8')
            )
            # TODO: Don't print here. Return this and let the caller print.
            logger.info("Spark online.")
        except Exception as e:
            # TODO: Catch a more specific problem known to be related to Spark not
            #       being up; provide a slightly better error message, and don't
            #       dump a large stack trace on the user.
            raise Exception("Spark health check failed.") from e



import concurrent.futures
import functools
import json
import os
import posixpath
import shlex
import sys
import logging
from concurrent.futures import FIRST_EXCEPTION

# External modules
import paramiko

# Flintrock modules
from .ssh import get_ssh_client, ssh_check_output, ssh, SSHKeyPair
from .exceptions import SSHError

FROZEN = getattr(sys, 'frozen', False)

if FROZEN:
    THIS_DIR = sys._MEIPASS
else:
    THIS_DIR = os.path.dirname(os.path.realpath(__file__))

SCRIPTS_DIR = os.path.join(THIS_DIR, 'scripts')


logger = logging.getLogger('flintrock.core')


class StorageDirs:
    def __init__(self, *, root, ephemeral, persistent):
        self.root = root
        self.ephemeral = ephemeral
        self.persistent = persistent


# TODO: Implement concept of ClusterNode. (?) That way we can
#       define a cluster as having several nodes, and implement
#       actions as `for node in nodes: node.action()`.
# NOTE: We take both IP addresses and host names because we
#       don't understand why Spark doesn't accept IP addresses
#       in its config, yet we prefer IP addresses when
#       connecting to hosts to avoid single-threaded DNS lookups.
#       See: https://github.com/nchammas/flintrock/issues/43
#       See: http://www.dalkescientific.com/writings/diary/archive/2012/01/19/concurrent.futures.html
class FlintrockCluster:
    def __init__(
            self,
            *,
            name,
            ssh_key_pair=None,
            storage_dirs=StorageDirs(root=None, ephemeral=None, persistent=None)):
        self.name = name
        self.ssh_key_pair = ssh_key_pair
        self.storage_dirs = storage_dirs
        self.services = []

    @property
    def master_ip(self) -> str:
        """
        The IP address of the master.

        Providers must override this property since it is typically derived from
        an underlying object, like an EC2 instance.
        """
        raise NotImplementedError

    @property
    def master_host(self) -> str:
        """
        The hostname of the master.

        Providers must override this property since it is typically derived from
        an underlying object, like an EC2 instance.
        """
        raise NotImplementedError

    @property
    def slave_ips(self) -> 'List[str]':
        """
        A list of the IP addresses of the slaves.

        Providers must override this property since it is typically derived from
        an underlying object, like an EC2 instance.
        """
        raise NotImplementedError

    @property
    def slave_hosts(self) -> 'List[str]':
        """
        A list of the hostnames of the slaves.

        Providers must override this property since it is typically derived from
        an underlying object, like an EC2 instance.
        """
        raise NotImplementedError

    @property
    def num_masters(self) -> int:
        """
        How many masters the cluster has.

        This normally just equals 1, but in cases where the cluster master
        has been destroyed this should return 0.

        Providers must override this property.
        """
        raise NotImplementedError

    @property
    def num_slaves(self) -> int:
        """
        How many slaves the cluster has.

        This is typically just len(self.slave_ips), but we need a separate
        property because slave IPs are not available when the cluster is
        stopped, and sometimes in that situation we still want to know how
        many slaves there are.

        Providers must override this property.
        """
        raise NotImplementedError

    def load_manifest(self, *, user: str, identity_file: str):
        """
        Load a cluster's manifest from the master. This will populate information
        about installed services and configured storage.

        Providers shouldn't need to override this method.
        """
        if not self.master_ip:
            return

        master_ssh_client = get_ssh_client(
            user=user,
            host=self.master_ip,
            identity_file=identity_file,
            wait=True,
            print_status=False)

        with master_ssh_client:
            manifest_raw = ssh_check_output(
                client=master_ssh_client,
                command="""
                    cat "$HOME/.flintrock-manifest.json"
                """)
            # TODO: Would it be better if storage (ephemeral and otherwise) was
            #       implemented as a Flintrock service and tracked in the manifest?
            ephemeral_dirs_raw = ssh_check_output(
                client=master_ssh_client,
                # It's generally safer to avoid using ls:
                # http://mywiki.wooledge.org/ParsingLs
                command="""
                    shopt -s nullglob
                    for f in /media/ephemeral*; do
                        echo "$f"
                    done
                """)

        manifest = json.loads(manifest_raw)

        self.ssh_key_pair = SSHKeyPair(
            public=manifest['ssh_key_pair']['public'],
            private=manifest['ssh_key_pair']['private'])

        services = []
        for [service_name, manifest] in manifest['services']:
            # TODO: Expose the classes being used here.
            service = globals()[service_name](**manifest)
            services.append(service)
        self.services = services

        storage_dirs = StorageDirs(
            root='/media/root',
            ephemeral=sorted(ephemeral_dirs_raw.splitlines()),
            persistent=None)
        self.storage_dirs = storage_dirs

    def destroy_check(self):
        """
        Check that the cluster is in a state in which it can be destroyed.

        Providers should override this method since we have no way to perform
        this check in a provider-agnostic way.
        """
        pass

    def destroy(self):
        """
        Destroy the cluster and any resources created specifically to support
        it.

        Providers should override this method since we have no way to destroy a
        cluster in a provider-agnostic way.

        Nonetheless, this method should be called before the underlying provider
        destroys the nodes. That way, if we ever add cleanup logic here to destroy
        resources external to the cluster it will get executed correctly.
        """
        pass

    def start_check(self):
        """
        Check that the cluster is in a state in which it can be started.

        The interface can use this method to decide whether it needs to prompt
        the user for confirmation. If the cluster cannot be started (e.g.
        because it's already running) then we don't want to show a prompt.

        Providers should override this method since we have no way to perform
        this check in a provider-agnostic way.
        """
        pass

    def start(self, *, user: str, identity_file: str):
        """
        Start up all the services installed on the cluster.

        This method assumes that the nodes constituting cluster were just
        started up by the provider (e.g. EC2, GCE, etc.) they're hosted on
        and are running.
        """
        self.load_manifest(user=user, identity_file=identity_file)

        partial_func = functools.partial(
            start_node,
            services=self.services,
            user=user,
            identity_file=identity_file,
            cluster=self)
        hosts = [self.master_ip] + self.slave_ips

        run_against_hosts(partial_func=partial_func, hosts=hosts)

        master_ssh_client = get_ssh_client(
            user=user,
            host=self.master_ip,
            identity_file=identity_file)

        with master_ssh_client:
            for service in self.services:
                service.configure_master(
                    ssh_client=master_ssh_client,
                    cluster=self)

        for service in self.services:
            service.health_check(master_host=self.master_ip)

    def stop_check(self):
        """
        Check that the cluster is in a state in which it can be stopped.

        Providers should override this method since we have no way to perform
        this check in a provider-agnostic way.
        """
        pass

    def stop(self):
        """
        Prepare the cluster to be stopped by the underlying provider.

        There's currently nothing to do here, but this method should be called
        before the underlying provider stops the nodes.
        """
        pass

    def add_slaves_check(self):
        pass

    def add_slaves(self, *, user: str, identity_file: str, new_hosts: list):
        """
        Add new slaves to the cluster.

        Providers should implement this with the following signature:

            add_slaves(self, *, user: str, identity_file: str, num_slaves: int, **provider_specific_options)

        This method should be called after the new hosts are online and have been
        added to the cluster's internal list.
        """
        hosts = [self.master_ip] + self.slave_ips
        partial_func = functools.partial(
            add_slaves_node,
            services=self.services,
            user=user,
            identity_file=identity_file,
            cluster=self,
            new_hosts=new_hosts)
        run_against_hosts(partial_func=partial_func, hosts=hosts)

        master_ssh_client = get_ssh_client(
            user=user,
            host=self.master_ip,
            identity_file=identity_file)
        with master_ssh_client:
            for service in self.services:
                service.configure_master(
                    ssh_client=master_ssh_client,
                    cluster=self)

    def remove_slaves(self, *, user: str, identity_file: str):
        """
        Remove some slaves from the cluster.

        Providers should implement this method with the following signature:

            remove_slaves(self, *, user: str, identity_file: str, num_slaves: int)

        This method should be called after the provider has removed the slaves
        from the cluster's internal list but before the instances themselves
        have been terminated.

        This method simply makes sure that the rest of the cluster knows that
        the relevant slaves are no longer part of the cluster.
        """
        self.load_manifest(user=user, identity_file=identity_file)

        partial_func = functools.partial(
            remove_slaves_node,
            user=user,
            identity_file=identity_file,
            services=self.services,
            cluster=self)
        hosts = [self.master_ip] + self.slave_ips

        run_against_hosts(partial_func=partial_func, hosts=hosts)

    def run_command_check(self):
        """
        Check that the cluster is in a state that supports running commands.

        Providers should override this method since we have no way to perform
        this check in a provider-agnostic way.
        """
        pass

    def run_command(
            self,
            *,
            master_only: bool,
            user: str,
            identity_file: str,
            command: tuple):
        """
        Run a shell command on each node of an existing cluster.

        If master_only is True, then run the comand on the master only.
        """
        if master_only:
            target_hosts = [self.master_ip]
        else:
            target_hosts = [self.master_ip] + self.slave_ips

        partial_func = functools.partial(
            run_command_node,
            user=user,
            identity_file=identity_file,
            command=command)
        hosts = target_hosts

        run_against_hosts(partial_func=partial_func, hosts=hosts)

    def copy_file_check(self):
        """
        Check that the cluster is in a state in which files can be copied to
        it.

        Providers should override this method since we have no way to perform
        this check in a provider-agnostic way.
        """
        pass

    def copy_file(
            self,
            *,
            master_only: bool,
            user: str,
            identity_file: str,
            local_path: str,
            remote_path: str):
        """
        Copy a file to each node of an existing cluster.

        If master_only is True, then copy the file to the master only.
        """
        if master_only:
            target_hosts = [self.master_ip]
        else:
            target_hosts = [self.master_ip] + self.slave_ips

        partial_func = functools.partial(
            copy_file_node,
            user=user,
            identity_file=identity_file,
            local_path=local_path,
            remote_path=remote_path)
        hosts = target_hosts

        run_against_hosts(partial_func=partial_func, hosts=hosts)

    def login(
            self,
            *,
            user: str,
            identity_file: str):
        """
        Interactively SSH into the cluster master.
        """
        ssh(
            host=self.master_ip,
            user=user,
            identity_file=identity_file)


def generate_template_mapping(
    *,
    cluster: FlintrockCluster,
    # If we add additional services later on we may want to refactor
    # this to take a list of services and dynamically pull the service
    # name.
    spark_executor_instances: int,
    hadoop_version: str,
    spark_version: str
) -> dict:
    """
    Generate a template mapping from a FlintrockCluster instance that we can use
    to fill in template parameters.
    """
    hadoop_root_dir = posixpath.join(cluster.storage_dirs.root, 'hadoop')
    hadoop_ephemeral_dirs = ','.join(
        posixpath.join(path, 'hadoop')
        for path in cluster.storage_dirs.ephemeral
    )
    spark_root_dir = posixpath.join(cluster.storage_dirs.root, 'spark')
    spark_ephemeral_dirs = ','.join(
        posixpath.join(path, 'spark')
        for path in cluster.storage_dirs.ephemeral
    )

    template_mapping = {
        'master_ip': cluster.master_ip,
        'master_host': cluster.master_host,
        'slave_ips': '\n'.join(cluster.slave_ips),
        'slave_hosts': '\n'.join(cluster.slave_hosts),

        'hadoop_version': hadoop_version,
        'hadoop_short_version': '.'.join(hadoop_version.split('.')[:2]),
        'spark_version': spark_version,
        'spark_short_version': '.'.join(spark_version.split('.')[:2]) if '.' in spark_version else spark_version,

        'spark_executor_instances': spark_executor_instances,

        'hadoop_root_dir': hadoop_root_dir,
        'hadoop_ephemeral_dirs': hadoop_ephemeral_dirs,
        'spark_root_dir': spark_root_dir,
        'spark_ephemeral_dirs': spark_ephemeral_dirs,

        # If ephemeral storage is available, it replaces the root volume, which is
        # typically persistent. We don't want to mix persistent and ephemeral
        # storage since that causes problems after cluster stop/start; some volumes
        # have leftover data, whereas others start fresh.
        'hadoop_root_ephemeral_dirs': hadoop_ephemeral_dirs if hadoop_ephemeral_dirs else hadoop_root_dir,
        'spark_root_ephemeral_dirs': spark_ephemeral_dirs if spark_ephemeral_dirs else spark_root_dir,
    }

    return template_mapping


# TODO: Cache these files. (?) They are being read potentially tens or
#       hundreds of times. Maybe it doesn't matter because the files
#       are so small.
def get_formatted_template(*, path: str, mapping: dict) -> str:
    with open(path) as f:
        formatted = f.read().format(**mapping)
    return formatted


def run_against_hosts(*, partial_func: functools.partial, hosts: list):
    """
    Run a function asynchronously against each of the provided hosts.

    This function assumes that partial_func accepts `host` as a keyword argument.
    """
    with concurrent.futures.ThreadPoolExecutor(len(hosts)) as executor:
        futures = {
            executor.submit(functools.partial(partial_func, host=host))
            for host in hosts
        }
        concurrent.futures.wait(futures, return_when=FIRST_EXCEPTION)
        for future in futures:
            future.result()


def get_java_major_version(client: paramiko.client.SSHClient):
    possible_cmds = [
        "$JAVA_HOME/bin/java -version",
        "java -version"
    ]

    for command in possible_cmds:
        try:
            output = ssh_check_output(
                client=client,
                command=command)
            tokens = output.split()
            # First line of the output is like: 'java version "1.8.0_20"'
            # Get the version string and strip out the first two parts of the
            # version as a tuple: (1, 8)
            if len(tokens) >= 3:
                version_parts = tokens[2].strip('"').split(".")
                if len(version_parts) >= 2:
                    return tuple(int(part) for part in version_parts[:2])
        except SSHError:
            pass

    return None


def ensure_java8(client: paramiko.client.SSHClient):
    host = client.get_transport().getpeername()[0]
    java_major_version = get_java_major_version(client)

    if not java_major_version or java_major_version < (1, 8):
        logger.info("[{h}] Installing Java 1.8...".format(h=host))

        ssh_check_output(
            client=client,
            command="""
                set -e

                # Install Java 1.8 first to protect packages that depend on Java from being removed.
                sudo yum install -y java-1.8.0-openjdk

                # Remove any older versions of Java to force the default Java to 1.8.
                # We don't use /etc/alternatives because it does not seem to update links in /usr/lib/jvm correctly,
                # and we don't just rely on JAVA_HOME because some programs use java directly in the PATH.
                sudo yum remove -y java-1.6.0-openjdk java-1.7.0-openjdk

                sudo sh -c "echo export JAVA_HOME=/usr/lib/jvm/jre >> /etc/environment"
                source /etc/environment
            """)


def setup_node(
        *,
        # Change this to take host, user, and identity_file?
        # Add some kind of caching for SSH connections so that they
        # can be looked up by host and reused?
        ssh_client: paramiko.client.SSHClient,
        services: list,
        cluster: FlintrockCluster):
    """
    Setup a new node.

    Cluster methods like provision_node() and add_slaves_node() should
    delegate the main work of setting up new nodes to this function.
    """
    host = ssh_client.get_transport().getpeername()[0]
    ssh_check_output(
        client=ssh_client,
        command="""
            set -e

            echo {private_key} > "$HOME/.ssh/id_rsa"
            echo {public_key} >> "$HOME/.ssh/authorized_keys"

            chmod 400 "$HOME/.ssh/id_rsa"
        """.format(
            private_key=shlex.quote(cluster.ssh_key_pair.private),
            public_key=shlex.quote(cluster.ssh_key_pair.public)))

    with ssh_client.open_sftp() as sftp:
        sftp.put(
            localpath=os.path.join(SCRIPTS_DIR, 'setup-ephemeral-storage.py'),
            remotepath='/tmp/setup-ephemeral-storage.py')

    logger.info("[{h}] Configuring ephemeral storage...".format(h=host))
    # TODO: Print some kind of warning if storage is large, since formatting
    #       will take several minutes (~4 minutes for 2TB).
    storage_dirs_raw = ssh_check_output(
        client=ssh_client,
        command="""
            set -e
            python /tmp/setup-ephemeral-storage.py
            rm -f /tmp/setup-ephemeral-storage.py
        """)
    storage_dirs = json.loads(storage_dirs_raw)

    cluster.storage_dirs.root = storage_dirs['root']
    cluster.storage_dirs.ephemeral = storage_dirs['ephemeral']

    ensure_java8(ssh_client)

    for service in services:
        try:
            service.install(
                ssh_client=ssh_client,
                cluster=cluster,
            )
        except Exception as e:
            raise Exception(
                "Failed to install {}."
                .format(type(service).__name__)
            ) from e


def provision_cluster(
        *,
        cluster: FlintrockCluster,
        services: list,
        user: str,
        identity_file: str):
    """
    Connect to a freshly launched cluster and install the specified services.
    """
    partial_func = functools.partial(
        provision_node,
        services=services,
        user=user,
        identity_file=identity_file,
        cluster=cluster)
    hosts = [cluster.master_ip] + cluster.slave_ips

    run_against_hosts(partial_func=partial_func, hosts=hosts)

    master_ssh_client = get_ssh_client(
        user=user,
        host=cluster.master_host,
        identity_file=identity_file)

    with master_ssh_client:
        manifest = {
            'services': [[type(m).__name__, m.manifest] for m in services],
            'ssh_key_pair': cluster.ssh_key_pair._asdict(),
        }
        # The manifest tells us how the cluster is configured. We'll need this
        # when we resize the cluster or restart it.
        ssh_check_output(
            client=master_ssh_client,
            command="""
                echo {m} > "$HOME/.flintrock-manifest.json"
                chmod go-rw "$HOME/.flintrock-manifest.json"
            """.format(
                m=shlex.quote(json.dumps(manifest, indent=4, sort_keys=True))
            ))

        for service in services:
            service.configure_master(
                ssh_client=master_ssh_client,
                cluster=cluster)

    for service in services:
        service.health_check(master_host=cluster.master_host)


def provision_node(
        *,
        services: list,
        user: str,
        host: str,
        identity_file: str,
        cluster: FlintrockCluster):
    """
    Connect to a freshly launched node, set it up for SSH access, configure ephemeral
    storage, and install the specified services.

    This method is role-agnostic; it runs on both the cluster master and slaves.
    This method is meant to be called asynchronously.
    """
    client = get_ssh_client(
        user=user,
        host=host,
        identity_file=identity_file,
        wait=True)

    with client:
        setup_node(
            ssh_client=client,
            services=services,
            cluster=cluster)
        for service in services:
            service.configure(
                ssh_client=client,
                cluster=cluster)


def start_node(
        *,
        services: list,
        user: str,
        host: str,
        identity_file: str,
        cluster: FlintrockCluster):
    """
    Connect to an existing node that has just been started up again and prepare it for
    work.

    This method is role-agnostic; it runs on both the cluster master and slaves.
    This method is meant to be called asynchronously.
    """
    ssh_client = get_ssh_client(
        user=user,
        host=host,
        identity_file=identity_file,
        wait=True)

    with ssh_client:
        # TODO: Consider consolidating ephemeral storage code under a dedicated
        #       Flintrock service.
        if cluster.storage_dirs.ephemeral:
            ssh_check_output(
                client=ssh_client,
                command="""
                    sudo chown "{u}:{u}" {d}
                """.format(
                    u=user,
                    d=' '.join(cluster.storage_dirs.ephemeral)))

        for service in services:
            service.configure(
                ssh_client=ssh_client,
                cluster=cluster)


def add_slaves_node(
        *,
        user: str,
        host: str,
        identity_file: str,
        services: list,
        cluster: FlintrockCluster,
        new_hosts: list):
    """
    If the node is new, set it up. If not, just reconfigure it to recognize
    the newly added nodes.

    This method is role-agnostic; it runs on both the cluster master and slaves.
    This method is meant to be called asynchronously.
    """
    is_new_host = host in new_hosts

    client = get_ssh_client(
        user=user,
        host=host,
        identity_file=identity_file,
        wait=is_new_host)

    with client:
        if is_new_host:
            setup_node(
                ssh_client=client,
                services=services,
                cluster=cluster)

        for service in services:
            service.configure(
                ssh_client=client,
                cluster=cluster)


def remove_slaves_node(
        *,
        user: str,
        host: str,
        identity_file: str,
        services: list,
        cluster: FlintrockCluster):
    """
    Update the services on a node to remove the provided slaves.

    This method is role-agnostic; it runs on both the cluster master and slaves.
    This method is meant to be called asynchronously.
    """
    ssh_client = get_ssh_client(
        user=user,
        host=host,
        identity_file=identity_file)

    for service in services:
        service.configure(
            ssh_client=ssh_client,
            cluster=cluster)


def run_command_node(*, user: str, host: str, identity_file: str, command: tuple):
    """
    Run a shell command on a node.

    This method is role-agnostic; it runs on both the cluster master and slaves.
    This method is meant to be called asynchronously.
    """
    ssh_client = get_ssh_client(
        user=user,
        host=host,
        identity_file=identity_file)

    logger.info("[{h}] Running command...".format(h=host))

    command_str = ' '.join(command)

    with ssh_client:
        ssh_check_output(
            client=ssh_client,
            command=command_str)

    logger.info("[{h}] Command complete.".format(h=host))


def copy_file_node(
        *,
        user: str,
        host: str,
        identity_file: str,
        local_path: str,
        remote_path: str):
    """
    Copy a file to the specified remote path on a node.

    This method is role-agnostic; it runs on both the cluster master and slaves.
    This method is meant to be called asynchronously.
    """
    ssh_client = get_ssh_client(
        user=user,
        host=host,
        identity_file=identity_file)

    with ssh_client:
        remote_dir = posixpath.dirname(remote_path)

        try:
            ssh_check_output(
                client=ssh_client,
                command="""
                    test -d {path}
                """.format(path=shlex.quote(remote_dir)))
        except Exception as e:
            # TODO: Catch more specific exception.
            raise Exception("Remote directory does not exist: {d}".format(d=remote_dir))

        with ssh_client.open_sftp() as sftp:
            logger.info("[{h}] Copying file...".format(h=host))

            sftp.put(localpath=local_path, remotepath=remote_path)

            logger.info("[{h}] Copy complete.".format(h=host))


# This is necessary down here since we have a circular import dependency between
# core.py and services.py. I've thought about how to remove this circular dependency,
# but for now this seems like what we need to go with.
# Flintrock modules
from .services import HDFS, Spark  # Used by start_cluster() # noqa
